from typing import Optional, Union


class ClientError(Exception):
    """
    Base class for all errors raised while attempting to communicate with StarkNet through Client.
    """

    def __init__(self, message: str, code: Optional[str] = None):
        self.code = code
        self.message = f"Client failed{f' with code {code}' if code is not None else ''}: {message}"
        super().__init__(self.message)


class ContractNotFoundError(ClientError):
    """
    Requested contract was not found.
    """

    def __init__(
        self,
        block_hash: Optional[Union[int, str]] = None,
        block_number: Optional[int] = None,
    ):
        require_block_identifier(block_hash, block_number)

        identifier = block_hash or block_number
        self.identifier = str(identifier) if isinstance(identifier, int) else identifier

        super().__init__(message=f"No contract found for identifier: {self.identifier}")


def require_block_identifier(block_hash, block_number):
    if block_hash is None and block_number is None:
        raise ValueError("One of block hash or number must be provided")



import warnings
from typing import Optional, List, Union, Dict
from dataclasses import replace

from starkware.crypto.signature.signature import get_random_private_key
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.services.api.feeder_gateway.feeder_gateway_client import (
    CastableToHash,
)

from starknet_py.net.client import Client
from starknet_py.net.client_models import (
    SentTransactionResponse,
    Hash,
    DeclaredContract,
    Tag,
    TransactionReceipt,
    BlockStateUpdate,
    StarknetBlock,
    Declare,
    Deploy,
    BlockTransactionTraces,
    EstimatedFee,
    Calls,
    TransactionStatus,
    DeployTransactionResponse,
    DeclareTransactionResponse,
)
from starknet_py.constants import FEE_CONTRACT_ADDRESS
from starknet_py.net.account.compiled_account_contract import COMPILED_ACCOUNT_CONTRACT
from starknet_py.net.models import (
    InvokeFunction,
    StarknetChainId,
    Transaction,
    BlockIdentifier,
    chain_from_network,
)
from starknet_py.net.networks import Network, MAINNET, TESTNET
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner, KeyPair
from starknet_py.net.signer import BaseSigner
from starknet_py.transactions.deploy import make_deploy_tx
from starknet_py.utils.crypto.facade import Call
from starknet_py.utils.data_transformer.execute_transformer import execute_transformer
from starknet_py.utils.sync import add_sync_methods
from starknet_py.net.models.address import AddressRepresentation, parse_address


@add_sync_methods
class AccountClient(Client):
    """
    Extends the functionality of :obj:`Client <starknet_py.net.client.Client>`,
    adding additional methods for creating the account contract.
    """

    # pylint: disable=too-many-public-methods
    def __init__(
        self,
        address: AddressRepresentation,
        client: Client,
        signer: Optional[BaseSigner] = None,
        key_pair: Optional[KeyPair] = None,
    ):
        """
        :param address: Address of the account contract
        :param client: Instance of GatewayClient which will be used to add transactions
        :param signer: Custom signer to be used by AccountClient.
                       If none is provieded, default
                       :py:class:`starknet_py.net.signer.stark_curve_signer.StarkCurveSigner` is used.
        :param key_pair: Key pair that will be used to create a default `Signer`
        """
        # pylint: disable=too-many-arguments
        if signer is None and key_pair is None:
            raise ValueError(
                "Either a signer or a key_pair must be provied in AccountClient constructor"
            )

        chain = chain_from_network(net=client.net, chain=client.chain)

        if chain is None and client is None:
            raise ValueError("One of chain or client must be provided")
        self.address = parse_address(address)
        self.client = client
        self.signer = signer or StarkCurveSigner(
            account_address=self.address, key_pair=key_pair, chain_id=self.client.chain
        )

    @property
    def chain(self) -> StarknetChainId:
        return self.client.chain

    @property
    def net(self) -> Network:
        return self.client.net

    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> StarknetBlock:
        return await self.client.get_block(
            block_hash=block_hash, block_number=block_number
        )

    async def get_block_traces(
        self,
        block_hash: [Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockTransactionTraces:
        return await self.client.get_block_traces(
            block_hash=block_hash, block_number=block_number
        )

    async def get_state_update(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockStateUpdate:
        return await self.client.get_state_update(
            block_hash=block_hash, block_number=block_number
        )

    async def get_storage_at(
        self, contract_address: Hash, key: int, block_hash: Union[Hash, Tag]
    ) -> int:
        return await self.client.get_storage_at(
            contract_address=contract_address, key=key, block_hash=block_hash
        )

    async def get_transaction(self, tx_hash: Hash) -> Transaction:
        return await self.client.get_transaction(tx_hash=tx_hash)

    async def get_transaction_receipt(self, tx_hash: Hash) -> TransactionReceipt:
        return await self.client.get_transaction_receipt(tx_hash=tx_hash)

    async def wait_for_tx(
        self,
        tx_hash: Hash,
        wait_for_accept: Optional[bool] = False,
        check_interval=5,
    ) -> (int, TransactionStatus):
        return await self.client.wait_for_tx(
            tx_hash=tx_hash,
            wait_for_accept=wait_for_accept,
            check_interval=check_interval,
        )

    async def call_contract(
        self, invoke_tx: InvokeFunction, block_hash: Union[Hash, Tag] = None
    ) -> List[int]:
        return await self.client.call_contract(
            invoke_tx=invoke_tx, block_hash=block_hash
        )

    async def get_class_hash_at(self, contract_address: Hash) -> int:
        return await self.client.get_class_hash_at(contract_address=contract_address)

    async def get_class_by_hash(self, class_hash: Hash) -> DeclaredContract:
        return await self.client.get_class_by_hash(class_hash=class_hash)

    async def _get_nonce(self) -> int:
        [nonce] = await self.call_contract(
            InvokeFunction(
                contract_address=self.address,
                entry_point_selector=get_selector_from_name("get_nonce"),
                calldata=[],
                signature=[],
                max_fee=0,
                version=0,
            ),
            block_hash="latest",
        )
        return nonce

    def _get_default_token_address(self) -> str:
        if self.net not in [TESTNET, MAINNET]:
            raise ValueError(
                "Token_address must be specified when using a custom net address"
            )

        return FEE_CONTRACT_ADDRESS

    async def get_balance(
        self, token_address: Optional[AddressRepresentation] = None
    ) -> int:
        """
        Checks account's balance of specified token.

        :param token_address: Address of the ERC20 contract.
                              If not specified it will be payment token (wrapped ETH) address.
        :return: Token balance
        """

        token_address = token_address or self._get_default_token_address()

        low, high = await self.call_contract(
            InvokeFunction(
                contract_address=parse_address(token_address),
                entry_point_selector=get_selector_from_name("balanceOf"),
                calldata=[self.address],
                signature=[],
                max_fee=0,
                version=0,
            )
        )

        return (high << 128) + low

    async def prepare_invoke_function(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: int = 0,
    ) -> InvokeFunction:
        """
        Takes calls and creates InvokeFunction from them

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version
        :return: InvokeFunction created from the calls (without the signature)
        """

        calls = calls if isinstance(calls, List) else [calls]

        nonce = await self._get_nonce()

        calldata_py = merge_calls(calls)
        calldata_py.append(nonce)

        wrapped_calldata, _ = execute_transformer.from_python(*calldata_py)

        transaction = InvokeFunction(
            entry_point_selector=get_selector_from_name("__execute__"),
            calldata=wrapped_calldata,
            contract_address=self.address,
            signature=[],
            max_fee=0,
            version=version,
        )

        max_fee = await self._get_max_fee(transaction, max_fee, auto_estimate)

        return InvokeFunction(
            entry_point_selector=get_selector_from_name("__execute__"),
            calldata=wrapped_calldata,
            contract_address=self.address,
            signature=[],
            max_fee=max_fee,
            version=version,
        )

    async def _get_max_fee(
        self,
        transaction: InvokeFunction,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ):
        if auto_estimate and max_fee is not None:
            raise ValueError(
                "Max_fee and auto_estimate are exclusive and cannot be provided at the same time."
            )

        if auto_estimate:
            estimate_fee = await self.estimate_fee(transaction)
            max_fee = int(estimate_fee.overall_fee * 1.1)

        if max_fee is None:
            raise ValueError("Max_fee must be specified when invoking a transaction")

        if max_fee == 0:
            warnings.warn(
                "Transaction will fail with max_fee set to 0. Change it to a higher value.",
                DeprecationWarning,
            )

        return max_fee

    async def sign_transaction(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: int = 0,
    ) -> InvokeFunction:
        """
        Takes calls and creates signed InvokeFunction

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version
        :return: InvokeFunction created from the calls
        """
        execute_tx = await self.prepare_invoke_function(
            calls, max_fee, auto_estimate, version
        )

        signature = self.signer.sign_transaction(execute_tx)
        execute_tx = add_signature_to_transaction(execute_tx, signature)

        return execute_tx

    async def send_transaction(
        self, transaction: InvokeFunction
    ) -> SentTransactionResponse:
        return await self.client.send_transaction(transaction=transaction)

    async def execute(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: int = 0,
    ) -> SentTransactionResponse:
        """
        Takes calls and executes transaction

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version
        :return: SentTransactionResponse
        """
        execute_transaction = await self.sign_transaction(
            calls, max_fee, auto_estimate, version
        )
        return await self.send_transaction(execute_transaction)

    async def deploy(self, transaction: Deploy) -> DeployTransactionResponse:
        return await self.client.deploy(transaction=transaction)

    async def declare(self, transaction: Declare) -> DeclareTransactionResponse:
        return await self.client.declare(transaction=transaction)

    async def estimate_fee(
        self,
        tx: InvokeFunction,
        block_hash: Optional[CastableToHash] = None,
        block_number: BlockIdentifier = "pending",
    ) -> EstimatedFee:
        """
        :param tx: Transaction which fee we want to calculate
        :param block_hash: Estimate fee at specific block hash
        :param block_number: Estimate fee at given block number (or "pending" for pending block)
        :return: Estimated fee
        """
        signature = self.signer.sign_transaction(tx)
        tx = add_signature_to_transaction(tx, signature)

        return await self.client.estimate_fee(
            tx=tx,
            block_hash=block_hash,
            block_number=block_number,
        )

    async def get_code(self, *args, **kwargs):
        return await self.client.get_code(*args, **kwargs)

    @staticmethod
    async def create_account(
        client: Client,
        private_key: Optional[int] = None,
        signer: Optional[BaseSigner] = None,
    ) -> "AccountClient":
        """
        Creates the account using
        `OpenZeppelin Account contract
        <https://github.com/starkware-libs/cairo-lang/blob/4e233516f52477ad158bc81a86ec2760471c1b65/src/starkware/starknet/third_party/open_zeppelin/Account.cairo>`_

        :param client: Instance of Client (i.e. FullNodeClient or GatewayClient)
                       which will be used to add the transactions
        :param private_key: Private Key used for the account
        :param signer: Signer used to create account and sign transaction
        :return: Instance of AccountClient which interacts with created account on given network
        """
        if signer is None:
            private_key = private_key or get_random_private_key()

            key_pair = KeyPair.from_private_key(private_key)
            address = await deploy_account_contract(client, key_pair.public_key)
            signer = StarkCurveSigner(
                account_address=address, key_pair=key_pair, chain_id=client.chain
            )
        else:
            address = await deploy_account_contract(client, signer.public_key)

        return AccountClient(
            client=client,
            address=address,
            signer=signer,
        )


async def deploy_account_contract(
    client: Client, public_key: int
) -> AddressRepresentation:
    deploy_tx = make_deploy_tx(
        constructor_calldata=[public_key],
        compiled_contract=COMPILED_ACCOUNT_CONTRACT,
    )
    result = await client.deploy(deploy_tx)
    await client.wait_for_tx(
        tx_hash=result.transaction_hash,
    )
    return result.contract_address


def add_signature_to_transaction(
    tx: InvokeFunction, signature: List[int]
) -> InvokeFunction:
    return replace(tx, signature=signature)


def merge_calls(calls: Calls) -> List:
    def parse_call(
        call: Call, current_data_len: int, entire_calldata: List
    ) -> (Dict, int, List):
        data = {
            "to": call.to_addr,
            "selector": call.selector,
            "data_offset": current_data_len,
            "data_len": len(call.calldata),
        }
        current_data_len += len(call.calldata)
        entire_calldata += call.calldata

        return data, current_data_len, entire_calldata

    calldata = []
    current_data_len = 0
    entire_calldata = []
    for call in calls:
        data, current_data_len, entire_calldata = parse_call(
            call, current_data_len, entire_calldata
        )
        calldata.append(data)

    return [calldata, entire_calldata]



from dataclasses import dataclass
from typing import List

from starkware.crypto.signature.signature import (
    private_to_stark_key,
)
from starkware.starknet.core.os.transaction_hash.transaction_hash import (
    calculate_transaction_hash_common,
    TransactionHashPrefix,
)

from starknet_py.net.models import (
    AddressRepresentation,
    StarknetChainId,
    Transaction,
    parse_address,
)
from starknet_py.net.signer.base_signer import BaseSigner
from starknet_py.utils.crypto.facade import message_signature


@dataclass
class KeyPair:
    private_key: int
    public_key: int

    @staticmethod
    def from_private_key(key: int) -> "KeyPair":
        return KeyPair(private_key=key, public_key=private_to_stark_key(key))


class StarkCurveSigner(BaseSigner):
    def __init__(
        self,
        account_address: AddressRepresentation,
        key_pair: KeyPair,
        chain_id: StarknetChainId,
    ):
        self.address = parse_address(account_address)
        self.key_pair = key_pair
        self.chain_id = chain_id

    @property
    def private_key(self) -> int:
        return self.key_pair.private_key

    @property
    def public_key(self) -> int:
        return self.key_pair.public_key

    def sign_transaction(
        self,
        transaction: Transaction,
    ) -> List[int]:
        tx_hash = calculate_transaction_hash_common(
            tx_hash_prefix=TransactionHashPrefix.INVOKE,
            version=transaction.version,
            contract_address=self.address,
            entry_point_selector=transaction.entry_point_selector,
            calldata=transaction.calldata,
            max_fee=transaction.max_fee,
            chain_id=self.chain_id.value,
            additional_data=[],
        )
        # pylint: disable=invalid-name
        r, s = message_signature(msg_hash=tx_hash, priv_key=self.private_key)
        return [r, s]



import typing
from typing import List, Optional, Union

import aiohttp
from marshmallow import EXCLUDE

from starknet_py.net.client import (
    Client,
)
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    SentTransactionResponse,
    TransactionReceipt,
    BlockStateUpdate,
    StarknetBlock,
    InvokeFunction,
    Hash,
    Tag,
    DeclaredContract,
    Transaction,
    Declare,
    Deploy,
    EstimatedFee,
    BlockTransactionTraces,
    DeclareTransactionResponse,
    DeployTransactionResponse,
)
from starknet_py.net.http_client import RpcHttpClient
from starknet_py.net.models import StarknetChainId, chain_from_network
from starknet_py.net.networks import Network
from starknet_py.net.rpc_schemas.rpc_schemas import (
    StarknetBlockSchema,
    BlockStateUpdateSchema,
    DeclaredContractSchema,
    TransactionReceiptSchema,
    TypesOfTransactionsSchema,
    SentTransactionSchema,
    DeclareTransactionResponseSchema,
    DeployTransactionResponseSchema,
)
from starknet_py.net.client_utils import convert_to_felt
from starknet_py.transaction_exceptions import TransactionNotReceivedError
from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class FullNodeClient(Client):
    def __init__(
        self,
        node_url: str,
        net: Network,
        chain: Optional[StarknetChainId] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Client for interacting with starknet json-rpc interface.

        :param node_url: Url of the node providing rpc interface
        :param net: StarkNet network identifier
        :param chain: Chain id of the network used by the rpc client
        :param session: Aiohttp session to be used for request. If not provided, client will create a session for
                        every request. When using a custom session, user is resposible for closing it manually.
        """
        self.url = node_url
        self._client = RpcHttpClient(url=node_url, session=session)
        self._chain = chain_from_network(net, chain)
        self._net = net

    @property
    def net(self) -> Network:
        return self._net

    @property
    def chain(self) -> StarknetChainId:
        return self._chain

    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> StarknetBlock:
        if block_hash is not None:
            res = await self._client.call(
                method_name="getBlockByHash",
                params={
                    "block_hash": convert_to_felt(block_hash),
                    "requested_scope": "FULL_TXNS",
                },
            )
            return StarknetBlockSchema().load(res, unknown=EXCLUDE)
        if block_number is not None:
            res = await self._client.call(
                method_name="getBlockByNumber",
                params={"block_number": block_number, "requested_scope": "FULL_TXNS"},
            )
            return StarknetBlockSchema().load(res, unknown=EXCLUDE)
        raise ValueError("Either block_hash or block_number is required")

    async def get_block_traces(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockTransactionTraces:
        raise NotImplementedError()

    async def get_state_update(
        self,
        block_hash: Union[Hash, Tag],
    ) -> BlockStateUpdate:
        res = await self._client.call(
            method_name="getStateUpdateByHash",
            params={"block_hash": convert_to_felt(block_hash)},
        )
        return BlockStateUpdateSchema().load(res, unknown=EXCLUDE)

    async def get_storage_at(
        self,
        contract_address: Hash,
        key: int,
        block_hash: Union[Hash, Tag],
    ) -> int:
        res = await self._client.call(
            method_name="getStorageAt",
            params={
                "contract_address": convert_to_felt(contract_address),
                "key": convert_to_felt(key),
                "block_hash": convert_to_felt(block_hash),
            },
        )
        res = typing.cast(str, res)
        return int(res, 16)

    async def get_transaction(
        self,
        tx_hash: Hash,
    ) -> Transaction:
        try:
            res = await self._client.call(
                method_name="getTransactionByHash",
                params={"transaction_hash": convert_to_felt(tx_hash)},
            )
        except ClientError as ex:
            raise TransactionNotReceivedError() from ex
        return TypesOfTransactionsSchema().load(res, unknown=EXCLUDE)

    async def get_transaction_by_block_hash(
        self, block_hash: Hash, index: int
    ) -> Transaction:
        """
        Get the details of transaction in block indentified block_hash and transaction index

        :param block_hash: Hash of the block
        :param index: Index of the transaction
        :return: Transaction object
        """
        res = await self._client.call(
            method_name="getTransactionByBlockHashAndIndex",
            params={
                "block_hash": convert_to_felt(block_hash),
                "index": index,
            },
        )
        return TypesOfTransactionsSchema().load(res, unknown=EXCLUDE)

    async def get_transaction_by_block_number(
        self, block_number: int, index: int
    ) -> Transaction:
        """
        Get the details of transaction in block indentified block number and transaction index

        :param block_number: Number of the block
        :param index: Index of the transaction
        :return: Transaction object
        """
        res = await self._client.call(
            method_name="getTransactionByBlockNumberAndIndex",
            params={
                "block_number": block_number,
                "index": index,
            },
        )
        return TypesOfTransactionsSchema().load(res, unknown=EXCLUDE)

    async def get_transaction_receipt(self, tx_hash: Hash) -> TransactionReceipt:
        res = await self._client.call(
            method_name="getTransactionReceipt",
            params={"transaction_hash": convert_to_felt(tx_hash)},
        )
        return TransactionReceiptSchema().load(res, unknown=EXCLUDE)

    async def estimate_fee(
        self,
        tx: InvokeFunction,
        block_hash: Union[Hash, Tag] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> EstimatedFee:
        raise NotImplementedError()

    async def call_contract(
        self,
        invoke_tx: InvokeFunction,
        block_hash: Union[Hash, Tag] = None,
    ) -> List[int]:
        if block_hash is None:
            raise ValueError(
                "block_hash is required for calls when using FullNodeClient"
            )

        res = await self._client.call(
            method_name="call",
            params={
                "contract_address": convert_to_felt(invoke_tx.contract_address),
                "entry_point_selector": convert_to_felt(invoke_tx.entry_point_selector),
                "calldata": [convert_to_felt(i) for i in invoke_tx.calldata],
                "block_hash": convert_to_felt(block_hash),
            },
        )
        return [int(i, 16) for i in res["result"]]

    async def send_transaction(
        self, transaction: InvokeFunction
    ) -> SentTransactionResponse:
        res = await self._client.call(
            method_name="addInvokeTransaction",
            params={
                "function_invocation": {
                    "contract_address": convert_to_felt(transaction.contract_address),
                    "entry_point_selector": convert_to_felt(
                        transaction.entry_point_selector
                    ),
                    "calldata": [convert_to_felt(i) for i in transaction.calldata],
                },
                "signature": [convert_to_felt(i) for i in transaction.signature],
                "max_fee": hex(transaction.max_fee),
                "version": hex(transaction.version),
            },
        )

        return SentTransactionSchema().load(res, unknown=EXCLUDE)

    async def deploy(self, transaction: Deploy) -> DeployTransactionResponse:
        contract_definition = transaction.dump()["contract_definition"]

        res = await self._client.call(
            method_name="addDeployTransaction",
            params={
                "contract_address_salt": transaction.contract_address_salt,
                "constructor_calldata": [
                    convert_to_felt(i) for i in transaction.constructor_calldata
                ],
                "contract_definition": {
                    "program": contract_definition["program"],
                    "entry_points_by_type": contract_definition["entry_points_by_type"],
                },
            },
        )

        return DeployTransactionResponseSchema().load(res, unknown=EXCLUDE)

    async def declare(self, transaction: Declare) -> DeclareTransactionResponse:
        contract_class = transaction.dump()["contract_class"]

        res = await self._client.call(
            method_name="addDeclareTransaction",
            params={
                "contract_class": {
                    "program": contract_class["program"],
                    "entry_points_by_type": contract_class["entry_points_by_type"],
                },
                "version": transaction.version,
            },
        )

        return DeclareTransactionResponseSchema().load(res, unknown=EXCLUDE)

    async def get_class_hash_at(self, contract_address: Hash) -> int:
        res = await self._client.call(
            method_name="getClassHashAt",
            params={"contract_address": convert_to_felt(contract_address)},
        )
        res = typing.cast(str, res)
        return int(res, 16)

    async def get_class_by_hash(self, class_hash: Hash) -> DeclaredContract:
        res = await self._client.call(
            method_name="getClass", params={"class_hash": convert_to_felt(class_hash)}
        )
        return DeclaredContractSchema().load(res, unknown=EXCLUDE)



import typing
from typing import Union, Optional, List

import aiohttp
from marshmallow import EXCLUDE

from starknet_py.net.client import Client
from starknet_py.net.client_models import (
    Transaction,
    SentTransactionResponse,
    ContractCode,
    BlockStateUpdate,
    StarknetBlock,
    InvokeFunction,
    StarknetTransaction,
    Hash,
    Tag,
    DeclaredContract,
    Declare,
    Deploy,
    TransactionStatusResponse,
    EstimatedFee,
    BlockTransactionTraces,
    DeployTransactionResponse,
    DeclareTransactionResponse,
    TransactionReceipt,
)
from starknet_py.net.gateway_schemas.gateway_schemas import (
    ContractCodeSchema,
    StarknetBlockSchema,
    SentTransactionSchema,
    BlockStateUpdateSchema,
    DeclaredContractSchema,
    TypesOfTransactionsSchema,
    TransactionStatusSchema,
    BlockTransactionTracesSchema,
    EstimatedFeeSchema,
    DeployTransactionResponseSchema,
    DeclareTransactionResponseSchema,
    TransactionReceiptSchema,
)
from starknet_py.net.http_client import GatewayHttpClient
from starknet_py.net.models import StarknetChainId, chain_from_network
from starknet_py.net.networks import Network, net_address_from_net
from starknet_py.net.client_errors import ContractNotFoundError
from starknet_py.net.client_utils import convert_to_felt, is_block_identifier
from starknet_py.transaction_exceptions import TransactionNotReceivedError
from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class GatewayClient(Client):
    def __init__(
        self,
        net: Network,
        chain: StarknetChainId = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Client for interacting with starknet gateway.

        :param net: Target network for the client. Can be a string with URL or one of ``"mainnet"``, ``"testnet"``
        :param chain: Chain used by the network. Required if you use a custom URL for ``net`` param.
        :param session: Aiohttp session to be used for request. If not provided, client will create a session for
                        every request. When using a custom session, user is resposible for closing it manually.
        """
        host = net_address_from_net(net)
        feeder_gateway_url = f"{host}/feeder_gateway"
        gateway_url = f"{host}/gateway"

        self._net = net
        self._chain = chain_from_network(net, chain)
        self._feeder_gateway_client = GatewayHttpClient(
            url=feeder_gateway_url, session=session
        )
        self._gateway_client = GatewayHttpClient(url=gateway_url, session=session)

    @property
    def chain(self) -> StarknetChainId:
        return self._chain

    @property
    def net(self) -> StarknetChainId:
        return self._net

    async def get_transaction(
        self,
        tx_hash: Hash,
    ) -> Transaction:
        res = await self._feeder_gateway_client.call(
            method_name="get_transaction",
            params={"transactionHash": convert_to_felt(tx_hash)},
        )

        if res["status"] in ("UNKNOWN", "NOT_RECEIVED"):
            raise TransactionNotReceivedError()

        return TypesOfTransactionsSchema().load(res["transaction"], unknown=EXCLUDE)

    async def get_transaction_status(
        self,
        tx_hash: Hash,
    ) -> TransactionStatusResponse:
        """
        Fetches the transaction's status and block number

        :param tx_hash: Transaction's hash representation
        :return: An object containing transaction's status and optional block hash, if transaction was accepted
        """
        res = await self._feeder_gateway_client.call(
            params={"transactionHash": convert_to_felt(tx_hash)},
            method_name="get_transaction_status",
        )
        if res["tx_status"] in ("UNKNOWN", "NOT_RECEIVED"):
            raise TransactionNotReceivedError()

        return TransactionStatusSchema().load(res)

    async def get_contract_addresses(self) -> dict:
        """
        Fetches the addresses of the StarkNet system contracts

        :return: A dictionary indexed with contract name and a value of contract's address
        """
        return await self._feeder_gateway_client.call(
            method_name="get_contract_addresses",
        )

    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> StarknetBlock:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._feeder_gateway_client.call(
            method_name="get_block", params=block_identifier
        )
        return StarknetBlockSchema().load(res, unknown=EXCLUDE)

    async def get_block_traces(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockTransactionTraces:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._feeder_gateway_client.call(
            method_name="get_block_traces", params=block_identifier
        )
        return BlockTransactionTracesSchema().load(res, unknown=EXCLUDE)

    async def get_state_update(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockStateUpdate:
        """
        Get the information about the result of executing the requested block

        :param block_hash: Block's hash
        :param block_number: Block's number
        :return: BlockStateUpdate oject representing changes in the requested block
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        res = await self._feeder_gateway_client.call(
            method_name="get_state_update",
            params=block_identifier,
        )
        return BlockStateUpdateSchema().load(res, unknown=EXCLUDE)

    async def get_storage_at(
        self,
        contract_address: Hash,
        key: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        :param contract_address: Contract's address on Starknet
        :param key: An address of the storage variable inside the contract.
        :param block_hash: Fetches the value of the variable at given block hash
        :param block_number: See above, uses block number instead of hash
        :return: Storage value of given contract
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._feeder_gateway_client.call(
            method_name="get_storage_at",
            params={
                **{
                    "contractAddress": convert_to_felt(contract_address),
                    "key": key,
                },
                **block_identifier,
            },
        )
        res = typing.cast(str, res)
        return int(res, 16)

    async def get_transaction_receipt(self, tx_hash: Hash) -> TransactionReceipt:
        res = await self._feeder_gateway_client.call(
            method_name="get_transaction_receipt",
            params={"transactionHash": convert_to_felt(tx_hash)},
        )
        return TransactionReceiptSchema().load(res, unknown=EXCLUDE)

    async def get_code(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> ContractCode:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        params = {
            **{"contractAddress": convert_to_felt(contract_address)},
            **block_identifier,
        }

        res = await self._feeder_gateway_client.call(
            method_name="get_code", params=params
        )

        if len(res["bytecode"]) == 0:
            raise ContractNotFoundError(
                f"No contract found with following identifier {block_identifier}"
            )

        return ContractCodeSchema().load(res, unknown=EXCLUDE)

    async def estimate_fee(
        self,
        tx: InvokeFunction,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> EstimatedFee:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        res = await self._feeder_gateway_client.post(
            method_name="estimate_fee",
            payload=InvokeFunction.Schema().dump(tx),
            params=block_identifier,
        )

        return EstimatedFeeSchema().load(res, unknown=EXCLUDE)

    async def call_contract(
        self,
        invoke_tx: InvokeFunction,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[int]:
        """
        Call the contract with given instance of InvokeTransaction

        :param invoke_tx: Invoke transaction
        :param block_hash: Block hash to execute the contract at specific point of time
        :param block_number: Block number (or "pending" for pending block) to execute the contract at
        :return: List of integers representing contract's function output (structured like calldata)
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._feeder_gateway_client.post(
            method_name="call_contract",
            params=block_identifier,
            payload=InvokeFunction.Schema().dump(invoke_tx),
        )

        return [int(v, 16) for v in res["result"]]

    async def send_transaction(
        self,
        transaction: InvokeFunction,
        token: Optional[str] = None,
    ) -> SentTransactionResponse:
        res = await self._add_transaction(transaction, token)
        return SentTransactionSchema().load(res, unknown=EXCLUDE)

    async def deploy(
        self,
        transaction: Deploy,
        token: Optional[str] = None,
    ) -> DeployTransactionResponse:
        res = await self._add_transaction(transaction, token)
        return DeployTransactionResponseSchema().load(res, unknown=EXCLUDE)

    async def declare(
        self,
        transaction: Declare,
        token: Optional[str] = None,
    ) -> DeclareTransactionResponse:
        res = await self._add_transaction(transaction, token)
        return DeclareTransactionResponseSchema().load(res, unknown=EXCLUDE)

    async def get_class_hash_at(self, contract_address: Hash) -> int:
        res = await self._feeder_gateway_client.call(
            method_name="get_class_hash_at",
            params={"contractAddress": convert_to_felt(contract_address)},
        )
        res = typing.cast(str, res)
        return int(res, 16)

    async def get_class_by_hash(self, class_hash: Hash) -> DeclaredContract:
        res = await self._feeder_gateway_client.call(
            method_name="get_class_by_hash",
            params={"classHash": convert_to_felt(class_hash)},
        )
        return DeclaredContractSchema().load(res, unknown=EXCLUDE)

    async def _add_transaction(
        self,
        tx: StarknetTransaction,
        token: Optional[str] = None,
    ) -> dict:
        res = await self._gateway_client.post(
            method_name="add_transaction",
            payload=StarknetTransaction.Schema().dump(obj=tx),
            params={"token": token} if token is not None else {},
        )
        return res


def get_block_identifier(
    block_hash: Optional[Union[Hash, Tag]] = None,
    block_number: Optional[Union[int, Tag]] = None,
) -> dict:
    if block_hash is not None and block_number is not None:
        raise ValueError(
            "Block_hash and block_number parameters are mutually exclusive."
        )

    if block_hash is not None:
        if is_block_identifier(block_hash):
            return {"blockNumber": block_hash}
        return {"blockHash": convert_to_felt(block_hash)}

    if block_number is not None:
        return {"blockNumber": block_number}

    return {}



from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import List, Any, Dict, Optional, Union, Iterable
from typing_extensions import Literal

from starkware.starknet.services.api.gateway.transaction import (
    InvokeFunction as IF,
    Transaction as T,
    ContractClass as CD,
    Deploy as D,
    Declare as DCL,
)


from starknet_py.utils.docs import as_our_module

InvokeFunction = as_our_module(IF)
StarknetTransaction = as_our_module(T)
ContractClass = as_our_module(CD)
Deploy = as_our_module(D)
Declare = as_our_module(DCL)


Hash = Union[int, str]
Tag = Literal["pending", "latest"]


@dataclass
class Call:
    to_addr: int
    selector: int
    calldata: List[int]


Calls = Union[Call, Iterable[Call]]


@dataclass
class Event:
    """
    Dataclass representing an event emited by transaction
    """

    from_address: int
    keys: List[int]
    data: List[int]


@dataclass
class L1toL2Message:
    """
    Dataclass representing a L1->L2 message
    """

    payload: List[int]
    l1_address: int
    l2_address: Optional[int] = None


@dataclass
class L2toL1Message:
    """
    Dataclass representing a L2->L1 message
    """

    payload: List[int]
    l1_address: int
    l2_address: Optional[int] = None


class TransactionType(Enum):
    """
    Enum representing transaction types
    """

    INVOKE = "INVOKE"
    DEPLOY = "DEPLOY"
    DECLARE = "DECLARE"


@dataclass
class Transaction(ABC):
    """
    Dataclass representing common attributes of all transactions
    """

    hash: int
    signature: List[int]
    max_fee: int

    def __post_init__(self):
        if self.__class__ == Transaction:
            raise TypeError("Cannot instantiate abstract Transaction class")


@dataclass
class InvokeTransaction(Transaction):
    """
    Dataclass representing invoke transaction
    """

    contract_address: int
    calldata: List[int]
    entry_point_selector: int


@dataclass
class DeclareTransaction(Transaction):
    """
    Dataclass representing declare transaction
    """

    class_hash: int
    sender_address: int


@dataclass
class DeployTransaction(Transaction):
    """
    Dataclass representing deploy transaction
    """

    contract_address: int
    constructor_calldata: List[int]
    # TODO add once RPC supports rpc transactions better
    # class_hash: Optional[int] = None


class TransactionStatus(Enum):
    """
    Enum representing transaction statuses
    """

    NOT_RECEIVED = "NOT_RECEIVED"
    RECEIVED = "RECEIVED"
    PENDING = "PENDING"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"
    ACCEPTED_ON_L1 = "ACCEPTED_ON_L1"
    REJECTED = "REJECTED"


@dataclass
class TransactionReceipt:
    """
    Dataclass representing details of sent transaction
    """

    # pylint: disable=too-many-instance-attributes

    hash: int
    status: TransactionStatus
    block_number: Optional[int] = None
    version: int = 0
    actual_fee: int = 0
    rejection_reason: Optional[str] = None

    events: List[Event] = None
    l2_to_l1_messages: List[L2toL1Message] = None
    l1_to_l2_consumed_message: Optional[L1toL2Message] = None


@dataclass
class SentTransactionResponse:
    """
    Dataclass representing a result of sending a transaction to starknet
    """

    transaction_hash: int
    code: Optional[str] = None


@dataclass
class DeclareTransactionResponse(SentTransactionResponse):
    """
    Dataclass representing a result of declaring a contract on starknet
    """

    class_hash: int = None


@dataclass
class DeployTransactionResponse(SentTransactionResponse):
    """
    Dataclass representing a result of deploying a contract to starknet
    """

    contract_address: int = None


class BlockStatus(Enum):
    """
    Enum representing block status
    """

    PENDING = "PENDING"
    REJECTED = "REJECTED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"
    ACCEPTED_ON_L1 = "ACCEPTED_ON_L1"
    PROVEN = "PROVEN"


@dataclass
class StarknetBlock:
    """
    Dataclass representing a transaction on starknet
    """

    # pylint: disable=too-many-instance-attributes

    block_hash: int
    parent_block_hash: int
    block_number: int
    status: BlockStatus
    root: int
    transactions: List[Transaction]
    timestamp: int


@dataclass
class BlockSingleTransactionTrace:
    function_invocation: dict
    signature: List[int]
    transaction_hash: int


@dataclass
class BlockTransactionTraces:
    traces: List[BlockSingleTransactionTrace]


@dataclass
class StorageDiff:
    address: int
    key: int
    value: int


@dataclass
class EstimatedFee:
    overall_fee: int
    gas_price: int
    gas_usage: int


@dataclass
class ContractDiff:
    address: int
    contract_hash: int


@dataclass
class BlockStateUpdate:
    """
    Dataclass representing a change in state of a block
    """

    block_hash: int
    new_root: int
    old_root: int
    storage_diffs: List[StorageDiff]
    contract_diffs: List[ContractDiff]
    declared_contracts: List[int]


@dataclass
class ContractCode:
    """
    Dataclass representing contract deployed to starknet
    """

    bytecode: List[int]
    abi: List[Dict[str, Any]]


@dataclass
class EntryPoint:
    """
    Dataclass representing contract entry point
    """

    offset: int
    selector: int


@dataclass
class EntryPointsByType:
    """
    Dataclass representing contract class entrypoints by entry point type
    """

    constructor: List[EntryPoint]
    external: List[EntryPoint]
    l1_handler: List[EntryPoint]


@dataclass
class DeclaredContract:
    """
    Dataclass representing contract declared to starknet
    """

    program: dict
    entry_points_by_type: EntryPointsByType


@dataclass
class TransactionStatusResponse:
    block_hash: Optional[int]
    transaction_status: TransactionStatus



import pytest

from marshmallow import Schema, ValidationError

from starknet_py.net.common_schemas.common_schemas import NonPrefixedHex
from starknet_py.net.rpc_schemas.rpc_schemas import Felt, StatusField, BlockStatusField
from starknet_py.net.client_models import TransactionStatus, BlockStatus


def test_serialize_felt():
    class SchemaWithFelt(Schema):
        value1 = Felt(data_key="value1")

    data = {"value1": 2137}

    serialized = SchemaWithFelt().dumps(data)
    assert '"value1": "0x859"' in serialized


def test_serialize_felt_throws_on_none():
    class SchemaWithFelt(Schema):
        value1 = Felt(data_key="value1")

    data = {"value1": None}
    with pytest.raises(TypeError):
        SchemaWithFelt().dumps(data)


def test_deserialize_felt():
    class SchemaWithFelt(Schema):
        value1 = Felt(data_key="value1")

    data = {"value1": "0x859"}

    deserialized = SchemaWithFelt().load(data)
    assert deserialized["value1"] == 2137


def test_deserialize_felt_throws_on_invalid_data():
    class SchemaWithFelt(Schema):
        value1 = Felt(data_key="value1")

    data = {"value1": "2137"}

    with pytest.raises(ValidationError) as exinfo:
        SchemaWithFelt().load(data)
    assert "Invalid value provided for felt" in str(exinfo.value)


def test_serialize_hex():
    class SchemaWithHex(Schema):
        value1 = NonPrefixedHex(data_key="value1")

    data = {"value1": 123}

    serialized = SchemaWithHex().dump(data)
    assert serialized["value1"] == "7b"


def test_deserialize_hex():
    class SchemaWithHex(Schema):
        value1 = NonPrefixedHex(data_key="value1")

    data = {"value1": "7b"}

    deserialized = SchemaWithHex().load(data)
    assert deserialized["value1"] == 123


def test_serialize_status_field():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": TransactionStatus.RECEIVED}

    serialized = SchemaWithStatusField().dumps(data)
    assert '"value1": "RECEIVED"' in serialized


def test_deserialize_status_field():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": "RECEIVED"}

    deserialized = SchemaWithStatusField().load(data)
    assert deserialized["value1"] == TransactionStatus.RECEIVED


def test_deserialize_status_field_throws_on_invalid_data():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": "SENT"}

    with pytest.raises(ValidationError) as exinfo:
        SchemaWithStatusField().load(data)

    assert "Invalid value provided for TransactionStatus" in str(exinfo.value)


def test_serialize_block_status_field():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": BlockStatus.PENDING}

    serialized = SchemaWithBlockStatusField().dumps(data)
    assert '"value1": "PENDING"' in serialized


def test_deserialize_block_status_field():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": "PENDING"}

    deserialized = SchemaWithBlockStatusField().load(data)
    assert deserialized["value1"] == BlockStatus.PENDING


def test_serialize_block_status_field_throws_on_invalid_data():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": "SENT"}

    with pytest.raises(ValidationError) as exinfo:
        SchemaWithBlockStatusField().load(data)

    assert "Invalid value for BlockStatus provided" in str(exinfo.value)



from typing import Union, Sequence

from starkware.starknet.public.abi import get_selector_from_name

# noinspection PyPep8Naming
from starkware.starknet.services.api.gateway.transaction import (
    InvokeFunction as IF,
    Deploy as D,
    Transaction as T,
    Declare as DCL,
)
from starkware.starknet.core.os.transaction_hash.transaction_hash import (
    calculate_transaction_hash_common,
    calculate_deploy_transaction_hash,
    TransactionHashPrefix,
)

# noinspection PyPep8Naming
from starkware.starknet.definitions.transaction_type import TransactionType as TT


from starknet_py.net.models.chains import StarknetChainId
from starknet_py.utils.crypto.facade import pedersen_hash
from starknet_py.utils.docs import as_our_module

InvokeFunction = as_our_module(IF)
Deploy = as_our_module(D)
Transaction = as_our_module(T)
TransactionType = as_our_module(TT)
Declare = as_our_module(DCL)


def compute_invoke_hash(
    contract_address: int,
    entry_point_selector: Union[int, str],
    calldata: Sequence[int],
    chain_id: StarknetChainId,
    max_fee: int,
    version: int,
) -> int:
    # pylint: disable=too-many-arguments
    """
    Computes invocation hash.

    :param contract_address: int
    :param entry_point_selector: Union[int, str]
    :param calldata: Sequence[int]
    :param chain_id: StarknetChainId
    :param max_fee: Max fee
    :param version: Contract version
    :return: calculated hash
    """
    if isinstance(entry_point_selector, str):
        entry_point_selector = get_selector_from_name(entry_point_selector)

    return calculate_transaction_hash_common(
        tx_hash_prefix=TransactionHashPrefix.INVOKE,
        contract_address=contract_address,
        entry_point_selector=entry_point_selector,
        calldata=calldata,
        chain_id=chain_id.value,
        hash_function=pedersen_hash,
        additional_data=[],
        max_fee=max_fee,
        version=version,
    )


def compute_deploy_hash(
    contract_address: int,
    calldata: Sequence[int],
    chain_id: StarknetChainId,
    version: int,
) -> int:
    """

    :param contract_address: int
    :param calldata: Sequence[int] (constructor arguments)
    :param chain_id: StarknetChainId
    :param version: Version of deploy transaction
    :return: calculated hash
    """
    return calculate_deploy_transaction_hash(
        contract_address=contract_address,
        constructor_calldata=calldata,
        chain_id=chain_id.value,
        hash_function=pedersen_hash,
        version=version,
    )



STARKNET_L1_ABI = [
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "from_address",
                "type": "uint256",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to_address",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256[]",
                "name": "payload",
                "type": "uint256[]",
            },
        ],
        "name": "ConsumedMessageToL1",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from_address",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "to_address",
                "type": "uint256",
            },
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "selector",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256[]",
                "name": "payload",
                "type": "uint256[]",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "nonce",
                "type": "uint256",
            },
        ],
        "name": "ConsumedMessageToL2",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "from_address",
                "type": "uint256",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to_address",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256[]",
                "name": "payload",
                "type": "uint256[]",
            },
        ],
        "name": "LogMessageToL1",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from_address",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "to_address",
                "type": "uint256",
            },
            {
                "indexed": True,
                "internalType": "uint256",
                "name": "selector",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256[]",
                "name": "payload",
                "type": "uint256[]",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "nonce",
                "type": "uint256",
            },
        ],
        "name": "LogMessageToL2",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "acceptedGovernor",
                "type": "address",
            }
        ],
        "name": "LogNewGovernorAccepted",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "nominatedGovernor",
                "type": "address",
            }
        ],
        "name": "LogNominatedGovernor",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [],
        "name": "LogNominationCancelled",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "operator",
                "type": "address",
            }
        ],
        "name": "LogOperatorAdded",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "operator",
                "type": "address",
            }
        ],
        "name": "LogOperatorRemoved",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "removedGovernor",
                "type": "address",
            }
        ],
        "name": "LogRemovedGovernor",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "bytes32",
                "name": "stateTransitionFact",
                "type": "bytes32",
            }
        ],
        "name": "LogStateTransitionFact",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "globalRoot",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "int256",
                "name": "blockNumber",
                "type": "int256",
            },
        ],
        "name": "LogStateUpdate",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "from_address", "type": "uint256"},
            {"internalType": "uint256[]", "name": "payload", "type": "uint256[]"},
        ],
        "name": "consumeMessageFromL2",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "finalize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "identify",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes", "name": "data", "type": "bytes"}],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "isFinalized",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "isFrozen",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "testedOperator", "type": "address"}
        ],
        "name": "isOperator",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "l1ToL2MessageNonce",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "msgHash", "type": "bytes32"}],
        "name": "l1ToL2Messages",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "msgHash", "type": "bytes32"}],
        "name": "l2ToL1Messages",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "programHash",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "newOperator", "type": "address"}
        ],
        "name": "registerOperator",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "to_address", "type": "uint256"},
            {"internalType": "uint256", "name": "selector", "type": "uint256"},
            {"internalType": "uint256[]", "name": "payload", "type": "uint256[]"},
        ],
        "name": "sendMessageToL2",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "newProgramHash", "type": "uint256"}
        ],
        "name": "setProgramHash",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "starknetAcceptGovernance",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "starknetCancelNomination",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "testGovernor", "type": "address"}
        ],
        "name": "starknetIsGovernor",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "newGovernor", "type": "address"}
        ],
        "name": "starknetNominateNewGovernor",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "governorForRemoval", "type": "address"}
        ],
        "name": "starknetRemoveGovernor",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "stateBlockNumber",
        "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "stateRoot",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "removedOperator", "type": "address"}
        ],
        "name": "unregisterOperator",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256[]", "name": "programOutput", "type": "uint256[]"},
            {"internalType": "uint256", "name": "onchainDataHash", "type": "uint256"},
            {"internalType": "uint256", "name": "onchainDataSize", "type": "uint256"},
        ],
        "name": "updateState",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]



from abc import ABC, abstractmethod

from starknet_py.constants import OZ_PROXY_STORAGE_KEY
from starknet_py.net.models import parse_address


# noinspection PyUnresolvedReferences
class ProxyCheck(ABC):
    @abstractmethod
    async def is_proxy(self, contract: "Contract") -> bool:
        """
        :return: `True` if contract is a proxy or `False` if it is not
        """

    @abstractmethod
    async def implementation_address(self, contract: "Contract") -> int:
        """
        :return: Address of contract being proxied by `contract` given as a parameter
        """


# noinspection PyUnresolvedReferences
class ArgentProxyCheck(ProxyCheck):
    async def is_proxy(self, contract: "Contract") -> bool:
        return "get_implementation" in contract.functions

    async def implementation_address(self, contract: "Contract") -> int:
        res = await contract.functions["get_implementation"].call()
        return res[0]


# noinspection PyUnresolvedReferences
class OpenZeppelinProxyCheck(ProxyCheck):
    def __init__(self):
        self.storage_key = OZ_PROXY_STORAGE_KEY

    async def is_proxy(self, contract: "Contract") -> bool:
        return await self.implementation_address(contract) != 0

    async def implementation_address(self, contract: "Contract") -> int:
        res = await contract.client.get_storage_at(
            contract_address=contract.address, key=self.storage_key
        )
        return parse_address(res)



from abc import abstractmethod, ABC
from contextlib import nullcontext
from enum import Enum
from typing import Optional

import aiohttp
from aiohttp import ClientSession, ClientResponse

from starknet_py.net.client_errors import ClientError


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"


class HttpClient(ABC):
    def __init__(self, url, session: Optional[aiohttp.ClientSession] = None):
        self.url = url
        self.session = session

    def http_session(self) -> ClientSession:
        if self.session is not None:
            # noinspection PyTypeChecker
            return nullcontext(self.session)
        return aiohttp.ClientSession()

    async def request(
        self,
        address: str,
        http_method: HttpMethod,
        params: Optional[dict] = None,
        payload: Optional[dict] = None,
    ):
        async with self.http_session() as session:
            return await self._make_request(
                session=session,
                address=address,
                http_method=http_method,
                params=params,
                payload=payload,
            )

    async def _make_request(
        self,
        session: aiohttp.ClientSession,
        address: str,
        http_method: HttpMethod,
        params: dict,
        payload: dict,
    ) -> dict:
        # pylint: disable=too-many-arguments
        async with session.request(
            method=http_method.value, url=address, params=params, json=payload
        ) as request:
            await self.handle_request_error(request)
            return await request.json(content_type=None)

    @abstractmethod
    async def handle_request_error(self, request: ClientResponse):
        """
        Handle an errors returned by make_request
        """


class GatewayHttpClient(HttpClient):
    async def call(self, method_name: str, params: Optional[dict] = None) -> dict:
        return await self.request(
            http_method=HttpMethod.GET, address=self.address(method_name), params=params
        )

    async def post(
        self, method_name: str, payload: dict, params: Optional[dict] = None
    ) -> dict:
        return await self.request(
            http_method=HttpMethod.POST,
            address=self.address(method_name),
            payload=payload,
            params=params,
        )

    def address(self, method_name):
        return f"{self.url}/{method_name}"

    async def handle_request_error(self, request: ClientResponse):
        await basic_error_handle(request)


class RpcHttpClient(HttpClient):
    async def call(self, method_name: str, params: dict) -> dict:
        payload = {
            "jsonrpc": "2.0",
            "method": f"starknet_{method_name}",
            "params": params,
            "id": 0,
        }

        result = await self.request(
            http_method=HttpMethod.POST, address=self.url, payload=payload
        )

        if "result" not in result:
            self.handle_rpc_error(result)
        return result["result"]

    @staticmethod
    def handle_rpc_error(result: dict):
        if "error" not in result:
            raise ServerError(body=result)
        raise ClientError(
            code=result["error"]["code"], message=result["error"]["message"]
        )

    async def handle_request_error(self, request: ClientResponse):
        await basic_error_handle(request)


async def basic_error_handle(request: ClientResponse):
    if request.status >= 300:
        raise ClientError(code=str(request.status), message=await request.text())


class ServerError(Exception):
    def __init__(self, body: dict):
        self.message = "Rpc request failed"
        self.body = body
        super().__init__(self.message)



# noinspection PyProtectedMember
import typing
from typing import Sequence


# noinspection PyProtectedMember
from web3._utils.contracts import prepare_transaction
from web3 import Web3
from web3.types import (
    BlockIdentifier as EthBlockIdentifier,
    ChecksumAddress,
    ABIFunction,
    ABIFunctionParams,
)

from starknet_py.net.models import StarknetChainId


def get_l1_starknet_contract_address(net: StarknetChainId):
    return {
        # Contract Proxy addresses
        StarknetChainId.TESTNET: "0xde29d060D45901Fb19ED6C6e959EB22d8626708e",
        StarknetChainId.MAINNET: "0xc662c410C0ECf747543f5bA90660f6ABeBD9C8c4",
    }[net]


class StarknetL1Contract:
    def __init__(
        self,
        net: StarknetChainId,
        web3: Web3,
    ):
        self.w3 = web3
        self.contract_address: ChecksumAddress = typing.cast(
            ChecksumAddress, get_l1_starknet_contract_address(net)
        )

    def l2_to_l1_messages(
        self, msg_hash: bytes, block_number: EthBlockIdentifier = None
    ) -> bytes:
        abi: ABIFunction = {
            "inputs": typing.cast(
                Sequence[ABIFunctionParams],
                [{"internalType": "bytes32", "name": "msgHash", "type": "bytes32"}],
            ),
            "name": "l2ToL1Messages",
            "outputs": typing.cast(
                Sequence[ABIFunctionParams],
                [{"internalType": "uint256", "name": "", "type": "uint256"}],
            ),
            "stateMutability": "view",
            "type": "function",
        }

        return self.w3.eth.call(
            prepare_transaction(
                address=typing.cast(ChecksumAddress, self.contract_address),
                w3=self.w3,
                transaction={"value": 0},
                fn_identifier="l2ToL1Messages",
                fn_abi=abi,
                fn_args=[msg_hash],
            ),
            block_number,
        )

    def l1_to_l2_messages(
        self, msg_hash: bytes, block_number: EthBlockIdentifier = None
    ) -> bytes:
        abi: ABIFunction = {
            "inputs": typing.cast(
                Sequence[ABIFunctionParams],
                [{"internalType": "bytes32", "name": "msgHash", "type": "bytes32"}],
            ),
            "name": "l1ToL2Messages",
            "outputs": typing.cast(
                Sequence[ABIFunctionParams],
                [{"internalType": "uint256", "name": "", "type": "uint256"}],
            ),
            "stateMutability": "view",
            "type": "function",
        }

        return self.w3.eth.call(
            prepare_transaction(
                address=self.contract_address,
                w3=self.w3,
                transaction={"value": 0},
                fn_identifier="l1ToL2Messages",
                fn_abi=abi,
                fn_args=[msg_hash],
            ),
            block_number,
        )



from starkware.starknet.services.api.feeder_gateway.response_objects import (
    TransactionStatus,
)
from starkware.starknet.public.abi import get_storage_var_address

TxStatus = TransactionStatus

ACCEPTED_STATUSES = (TxStatus.ACCEPTED_ON_L1, TxStatus.ACCEPTED_ON_L2)

# Address came from starkware-libs/starknet-addresses repository: https://github.com/starkware-libs/starknet-addresses
FEE_CONTRACT_ADDRESS = (
    "0x049d36570d4e46f48e99674bd3fcc84644ddd6b96f7c741b1562b82f9e004dc7"
)

OZ_PROXY_STORAGE_KEY = get_storage_var_address("Proxy_implementation_address")



class TransactionFailedError(Exception):
    """
    Base exception for transaction failure
    """

    def __init__(
        self,
        message: str = "Unknown starknet error",
    ):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"Transaction failed with following starknet error: {self.message}."


class TransactionRejectedError(TransactionFailedError):
    """
    Exception for transactions rejected by starknet
    """

    def __str__(self):
        return (
            "Transaction was rejected with following starknet error: "
            f"{self.message}."
        )


class TransactionNotReceivedError(TransactionFailedError):
    """
    Exception for transactions not received on starknet
    """

    def __init__(self):
        super().__init__(message="Transaction not received")

    def __str__(self):
        return "Transaction was not received on starknet"



from typing import Optional, List

from starkware.starknet.services.api.contract_class import ContractClass

from starknet_py.compile.compiler import (
    StarknetCompilationSource,
    Compiler,
    create_contract_class,
)


def create_compiled_contract(
    compilation_source: Optional[StarknetCompilationSource] = None,
    compiled_contract: Optional[str] = None,
    search_paths: Optional[List[str]] = None,
) -> ContractClass:
    if not compiled_contract and not compilation_source:
        raise ValueError("One of compiled_contract or compilation_source is required.")

    if not compiled_contract:
        compiled_contract = Compiler(
            contract_source=compilation_source, cairo_path=search_paths
        ).compile_contract()
    definition = create_contract_class(compiled_contract)
    return definition



import inspect
from functools import wraps
from typing import TypeVar

from asgiref.sync import async_to_sync

T = TypeVar("T")


def make_sync(fn):
    sync_fun = async_to_sync(fn)

    @wraps(fn)
    def impl(*args, **kwargs):
        return sync_fun(*args, **kwargs)

    return impl


def add_sync_methods(original_class: T) -> T:
    """
    Decorator for adding a synchronous version of a class.
    :param original_class: Input class
    :return: Input class with .sync property that contains synchronous version of this class
    """
    properties = {**original_class.__dict__}
    for name, value in properties.items():
        sync_name = name + "_sync"

        # Hand written implementation exists
        if sync_name in properties:
            continue

        # Make all callables synchronous
        if inspect.iscoroutinefunction(value):
            setattr(original_class, sync_name, make_sync(value))
        elif isinstance(value, staticmethod) and inspect.iscoroutinefunction(
            value.__func__
        ):
            setattr(original_class, sync_name, staticmethod(make_sync(value.__func__)))
        elif isinstance(value, classmethod) and inspect.iscoroutinefunction(
            value.__func__
        ):
            setattr(original_class, sync_name, classmethod(make_sync(value.__func__)))

    return original_class



from marshmallow import Schema, fields, post_load, EXCLUDE
from marshmallow_oneofschema import OneOfSchema

from starknet_py.net.client_models import (
    ContractCode,
    StarknetBlock,
    L2toL1Message,
    L1toL2Message,
    SentTransactionResponse,
    ContractDiff,
    StorageDiff,
    BlockStateUpdate,
    EntryPoint,
    EntryPointsByType,
    DeclaredContract,
    InvokeTransaction,
    DeployTransaction,
    DeclareTransaction,
    TransactionReceipt,
    TransactionStatusResponse,
    BlockTransactionTraces,
    BlockSingleTransactionTrace,
    EstimatedFee,
    Event,
    DeclareTransactionResponse,
    DeployTransactionResponse,
)
from starknet_py.net.common_schemas.common_schemas import (
    Felt,
    BlockStatusField,
    StatusField,
    NonPrefixedHex,
)

# pylint: disable=unused-argument
# pylint: disable=no-self-use


class EventSchema(Schema):
    from_address = Felt(data_key="from_address", required=True)
    keys = fields.List(Felt(), data_key="keys", required=True)
    data = fields.List(Felt(), data_key="data", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Event:
        return Event(**data)


class L1toL2MessageSchema(Schema):
    l1_address = Felt(data_key="from_address", required=True)
    l2_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L1toL2Message:
        return L1toL2Message(**data)


class L2toL1MessageSchema(Schema):
    l2_address = Felt(data_key="from_address", required=True)
    l1_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L2toL1Message:
        return L2toL1Message(**data)


class TransactionSchema(Schema):
    hash = Felt(data_key="transaction_hash", required=True)
    signature = fields.List(Felt(), data_key="signature", load_default=[])
    max_fee = Felt(load_default=0)


class InvokeTransactionSchema(TransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> InvokeTransaction:
        return InvokeTransaction(**data)


class DeployTransactionSchema(TransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)
    constructor_calldata = fields.List(
        Felt(), data_key="constructor_calldata", required=True
    )
    # class_hash = Felt(data_key="class_hash", load_default=0)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployTransaction:
        return DeployTransaction(**data)


class DeclareTransactionSchema(TransactionSchema):
    class_hash = Felt(data_key="class_hash", required=True)
    sender_address = Felt(data_key="sender_address", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransaction:
        return DeclareTransaction(**data)


class TypesOfTransactionsSchema(OneOfSchema):
    type_field = "type"
    type_schemas = {
        "INVOKE_FUNCTION": InvokeTransactionSchema,
        "DECLARE": DeclareTransactionSchema,
        "DEPLOY": DeployTransactionSchema,
    }


class TransactionReceiptSchema(Schema):
    hash = Felt(data_key="transaction_hash", required=True)
    status = StatusField(data_key="status", required=True)
    block_number = fields.Integer(data_key="block_number", load_default=None)
    version = fields.Integer(data_key="version", allow_none=True)
    actual_fee = Felt(data_key="actual_fee", allow_none=True)
    rejection_reason = fields.Dict(
        keys=fields.String(),
        values=fields.Raw(),
        data_key="transaction_failure_reason",
        allow_none=True,
        default=None,
    )
    events = fields.List(
        fields.Nested(EventSchema()), data_key="events", load_default=[]
    )
    l1_to_l2_consumed_message = fields.Nested(
        L1toL2MessageSchema(), data_key="l1_to_l2_consumed_message", load_default=None
    )
    l2_to_l1_messages = fields.List(
        fields.Nested(L2toL1MessageSchema()),
        data_key="l2_to_l1_messages",
        load_default=[],
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionReceipt:
        if data.get("rejection_reason", None) is not None:
            rejection_reason = data["rejection_reason"]["error_message"]
            del data["rejection_reason"]

            return TransactionReceipt(**data, rejection_reason=rejection_reason)

        return TransactionReceipt(**data)


class ContractCodeSchema(Schema):
    bytecode = fields.List(Felt(), data_key="bytecode", required=True)
    abi = fields.List(
        fields.Dict(keys=fields.String(), values=fields.Raw()),
        data_key="abi",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs):
        return ContractCode(**data)


class StarknetBlockSchema(Schema):
    block_hash = Felt(data_key="block_hash", required=True)
    parent_block_hash = Felt(data_key="parent_block_hash", required=True)
    block_number = fields.Integer(data_key="block_number", required=True)
    status = BlockStatusField(data_key="status", required=True)
    root = NonPrefixedHex(data_key="state_root", required=True)
    transactions = fields.List(
        fields.Nested(TypesOfTransactionsSchema(unknown=EXCLUDE)),
        data_key="transactions",
        required=True,
    )
    timestamp = fields.Integer(data_key="timestamp", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return StarknetBlock(**data)


class BlockSingleTransactionTraceSchema(Schema):
    function_invocation = fields.Dict(
        keys=fields.String(),
        values=fields.Raw(),
        data_key="function_invocation",
        required=True,
    )
    signature = fields.List(Felt(), data_key="signature", load_default=[])
    transaction_hash = Felt(data_key="transaction_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return BlockSingleTransactionTrace(**data)


class BlockTransactionTracesSchema(Schema):
    traces = fields.List(
        fields.Nested(BlockSingleTransactionTraceSchema(unknown=EXCLUDE)),
        data_key="traces",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs):
        return BlockTransactionTraces(**data)


class EstimatedFeeSchema(Schema):
    overall_fee = fields.Integer(data_key="overall_fee", required=True)
    gas_price = fields.Integer(data_key="gas_price", required=True)
    gas_usage = fields.Integer(data_key="gas_usage", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return EstimatedFee(**data)


class SentTransactionSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)
    code = fields.String(data_key="code", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return SentTransactionResponse(**data)


class DeclareTransactionResponseSchema(SentTransactionSchema):
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return DeclareTransactionResponse(**data)


class DeployTransactionResponseSchema(SentTransactionSchema):
    contract_address = Felt(data_key="address", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return DeployTransactionResponse(**data)


class StorageDiffSchema(Schema):
    address = Felt(data_key="address", required=True)
    key = Felt(data_key="key", required=True)
    value = Felt(data_key="value", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return StorageDiff(**data)


class ContractDiffsSchema(Schema):
    address = Felt(data_key="address", required=True)
    contract_hash = NonPrefixedHex(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return ContractDiff(**data)


class BlockStateUpdateSchema(Schema):
    block_hash = Felt(data_key="block_hash", required=True)
    new_root = NonPrefixedHex(data_key="new_root", required=True)
    old_root = NonPrefixedHex(data_key="old_root", required=True)
    state_diff = fields.Dict(
        keys=fields.String(), values=fields.Raw(), data_key="state_diff", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs):
        contracts_diffs = data["state_diff"]["deployed_contracts"]
        contracts_diffs = [
            ContractDiffsSchema().load(contract) for contract in contracts_diffs
        ]

        storage_diffs = []
        for address, diffs in data["state_diff"]["storage_diffs"].items():
            for diff in diffs:
                storage_diff = StorageDiffSchema().load(
                    {"address": address, "key": diff["key"], "value": diff["value"]}
                )
                storage_diffs.append(storage_diff)

        declared_contracts = data["state_diff"]["declared_contracts"]
        declared_contracts = [
            int(declared_contract, 16) for declared_contract in declared_contracts
        ]

        del data["state_diff"]

        return BlockStateUpdate(
            **data,
            storage_diffs=storage_diffs,
            contract_diffs=contracts_diffs,
            declared_contracts=declared_contracts,
        )


class EntryPointSchema(Schema):
    offset = Felt(data_key="offset", required=True)
    selector = Felt(data_key="selector", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPoint:
        return EntryPoint(**data)


class EntryPointsByTypeSchema(Schema):
    constructor = fields.List(
        fields.Nested(EntryPointSchema()), data_key="CONSTRUCTOR", required=True
    )
    external = fields.List(
        fields.Nested(EntryPointSchema()), data_key="EXTERNAL", required=True
    )
    l1_handler = fields.List(
        fields.Nested(EntryPointSchema()), data_key="L1_HANDLER", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPointsByType:
        return EntryPointsByType(**data)


class DeclaredContractSchema(Schema):
    program = fields.Dict(
        keys=fields.String(),
        values=fields.Raw(allow_none=True),
        data_key="program",
        required=True,
    )
    entry_points_by_type = fields.Nested(
        EntryPointsByTypeSchema(), data_key="entry_points_by_type", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclaredContract:
        return DeclaredContract(**data)


class TransactionStatusSchema(Schema):
    transaction_status = StatusField(data_key="tx_status", required=True)
    block_hash = fields.Integer(data_key="block_hash", allow_none=True)

    @post_load
    def make_result(self, data, **kwargs) -> TransactionStatusResponse:
        return TransactionStatusResponse(**data)



# pylint: disable=unused-import
# Backward compatibility
# noinspection PyUnresolvedReferences
from starknet_py.net.models import (
    InvokeFunction,
    Deploy,
    Transaction,
)

# noinspection PyUnresolvedReferences
from starknet_py.net.models.address import parse_address

# noinspection PyUnresolvedReferences
from starknet_py.net.networks import net_address_from_net

# noinspection PyUnresolvedReferences
from starknet_py.cairo.felt import encode_shortstring, decode_shortstring



from typing import Union, Sequence

from starkware.cairo.common.hash_state import compute_hash_on_elements
from starkware.starknet.core.os.contract_address.contract_address import (
    CONTRACT_ADDRESS_PREFIX,
)
from starkware.starknet.services.api.feeder_gateway.feeder_gateway_client import (
    BlockIdentifier,
)

from starknet_py.utils.crypto.facade import pedersen_hash
from starknet_py.utils.docs import as_our_module

AddressRepresentation = Union[int, str]
Address = int


def parse_address(value: AddressRepresentation) -> Address:
    if isinstance(value, int):
        return value

    try:
        return int(value, 16)
    except TypeError as t_err:
        raise TypeError("Invalid address format.") from t_err


def compute_address(
    contract_hash: int, constructor_calldata: Sequence[int], salt: int
) -> int:
    """
    Computes contract's address.

    :param contract_hash: int
    :param constructor_calldata: Sequence[int]
    :param salt: int
    :return: Contract's address
    """
    constructor_calldata_hash = compute_hash_on_elements(
        data=constructor_calldata, hash_func=pedersen_hash
    )
    caller_address = 0
    return compute_hash_on_elements(
        data=[
            CONTRACT_ADDRESS_PREFIX,
            caller_address,
            salt,
            contract_hash,
            constructor_calldata_hash,
        ],
        hash_func=pedersen_hash,
    )


BlockIdentifier = as_our_module(BlockIdentifier)



from typing import Union

MAINNET = "mainnet"
TESTNET = "testnet"

try:
    # noinspection PyUnresolvedReferences
    from typing import Literal  # pylint: disable=no-name-in-module

    PredefinedNetwork = Literal["mainnet", "testnet"]
except ImportError:
    PredefinedNetwork = str

Network = Union[PredefinedNetwork, str]


def net_address_from_net(net: Network) -> str:
    return {
        MAINNET: "https://alpha-mainnet.starknet.io",
        TESTNET: "https://alpha4.starknet.io",
    }.get(net, net)



from typing import Optional

from starkware.starknet.definitions.general_config import (
    StarknetChainId as _StarknetChainId,
)

from starknet_py.net.networks import Network, MAINNET, TESTNET
from starknet_py.utils.docs import as_our_module

StarknetChainId = as_our_module(_StarknetChainId)


def chain_from_network(
    net: Network, chain: Optional[StarknetChainId]
) -> StarknetChainId:
    mapping = {
        MAINNET: StarknetChainId.MAINNET,
        TESTNET: StarknetChainId.TESTNET,
    }

    if net in mapping:
        return mapping[net]

    if not chain:
        raise ValueError("Chain is required when not using predefined networks.")

    return chain



from dataclasses import dataclass
from functools import reduce
from typing import List, Optional

from eth_utils import keccak
from starkware.starknet.services.api.feeder_gateway.response_objects import (
    TransactionReceipt,
)
from web3 import Web3
from web3.types import BlockIdentifier as EthBlockIdentifier

from starknet_py.net.client import Client
from starknet_py.net.l1.contracts import StarknetL1Contract
from starknet_py.net.l1.starknet_l1_abi import STARKNET_L1_ABI
from starknet_py.net.models import (
    StarknetChainId,
    AddressRepresentation,
    parse_address,
)

from starknet_py.utils.sync import add_sync_methods


def encode_packed(*args: int) -> bytes:
    # noinspection PyUnresolvedReferences
    return reduce(
        lambda acc, x: acc + x,
        [x.to_bytes(32, byteorder="big", signed=False) for x in args],
    )


@dataclass
class MessageToEthContent:
    """A dataclass representing a message's payload (from Starknet to Ethereum)"""

    starknet_sender: AddressRepresentation
    eth_recipient: int  # Integer representation of l1 hex address
    payload: List[int]

    @property
    def hash(self):
        return keccak(
            encode_packed(
                parse_address(self.starknet_sender),
                self.eth_recipient,
                len(self.payload),
                *self.payload,
            )
        )


def int_from_bytes(byte: bytes) -> int:
    return int(byte.hex(), 16)


@add_sync_methods
@dataclass
class MessageToEth:
    """

    Class which enables you to check Starknet to Ethereum messages status
    """

    hash: bytes  # 32 bytes hash representation

    @classmethod
    def from_hash(cls, msg_hash: bytes) -> "MessageToEth":
        """

        :param msg_hash: Message's hash
        :return: Instance of `MessageToEth`
        """
        return cls(hash=msg_hash)

    @classmethod
    def from_content(cls, msg_content: MessageToEthContent) -> "MessageToEth":
        """

        :param msg_content: Dataclass representing message's content
        :return: Instance of `MessageToEth`
        """
        return cls.from_hash(msg_content.hash)

    @classmethod
    def from_tx_receipt(cls, tx_receipt: TransactionReceipt) -> List["MessageToEth"]:
        """
        :param tx_receipt: A JSON from starknet.py's Client `get_transaction_receipt`
        :return: A list of Starknet to Ethereum messages
        """
        return [
            MessageToEth.from_content(
                MessageToEthContent(
                    starknet_sender=message.from_address,
                    eth_recipient=int(message.to_address, 16),
                    payload=[int(felt_str) for felt_str in message.payload],
                )
            )
            for message in tx_receipt.l2_to_l1_messages
        ]

    @classmethod
    async def from_tx_hash(
        cls,
        tx_hash: str,
        client: Client,
    ) -> List["MessageToEth"]:
        """

        :param tx_hash: A starknet transaction hash
        :param client: Instance of starknet.py's Client class
        :return: A list of messages to Ethereum in this transaction
        """
        receipt = await client.get_transaction_receipt(tx_hash)
        return cls.from_tx_receipt(receipt)

    def count_queued_sync(
        self,
        chain_id: StarknetChainId,
        web3: Web3,
        block_number: Optional[EthBlockIdentifier] = None,
    ) -> int:
        """

        :param chain_id: A `StarknetChainId` (which contains StarkNet core contract deployed)
        :param web3: Web3 instance from web3.py
        :param block_number: Optional. `EthBlockIdentifier` which is a hex address, integer,
                             or one of "latest", "earliest", "pending"
        :return: an integer (ranging from 0 upwards, representing the number of messages on L1 waiting for consumption)
        """
        return int_from_bytes(
            StarknetL1Contract(chain_id, web3).l2_to_l1_messages(
                self.hash, block_number
            )
        )


@dataclass
class MessageToStarknetContent:
    """A dataclass representing a Ethereum to Starknet message payload (from L1 to L2)"""

    eth_sender: int  # Integer representation of l1 hex address
    starknet_recipient: AddressRepresentation
    nonce: int
    selector: int
    payload: List[int]

    @property
    def hash(self):
        return keccak(
            encode_packed(
                self.eth_sender,
                parse_address(self.starknet_recipient),
                self.nonce,
                self.selector,
                len(self.payload),
                *self.payload,
            )
        )

    @classmethod
    def from_receipt(cls, receipt, web3: Web3) -> List["MessageToStarknetContent"]:
        logs = (
            web3.eth.contract(abi=STARKNET_L1_ABI)
            .events.LogMessageToL2()
            .processReceipt(receipt)
        )
        return [
            cls(
                eth_sender=int(log.args["from_address"], 16),
                starknet_recipient=log.args["to_address"],
                nonce=log.args["nonce"],
                selector=log.args["selector"],
                payload=log.args["payload"],
            )
            for log in logs
        ]


@add_sync_methods
@dataclass
class MessageToStarknet:
    """

    Class which enables you to check Eth -> Starknet (L1 -> L2) messages status
    """

    hash: bytes  # 32 bytes hash representation

    @classmethod
    def from_hash(cls, msg_hash: bytes) -> "MessageToStarknet":
        """

        :param msg_hash: Message's hash
        :return: Instance of `MessageToStarknet`
        """
        return cls(hash=msg_hash)

    @classmethod
    def from_content(cls, msg_content: MessageToStarknetContent) -> "MessageToStarknet":
        """

        :param msg_content: Dataclass representing message's content
        :return: Instance of `MessageToStarknet`
        """
        return cls.from_hash(msg_content.hash)

    @classmethod
    def from_tx_receipt(cls, receipt, web3: Web3) -> List["MessageToStarknet"]:
        """

        :param receipt: Transaction receipt object from web3.py
        :param web3: Web3 instance from web3.py
        :return: A list of message to Starknet (L1 to L2) messages in this receipt
        """
        return [
            cls.from_content(msg_content)
            for msg_content in MessageToStarknetContent.from_receipt(receipt, web3)
        ]

    @classmethod
    async def from_tx_hash(cls, tx_hash: str, web3: Web3) -> List["MessageToStarknet"]:
        """

        :param tx_hash: Transaction hash including some L1 to L2 messages
        :param web3: Web3 instance from web3.py
        :return: A list of messages to StarkNet in this transaction
        """
        receipt = web3.eth.getTransactionReceipt(tx_hash)
        return cls.from_tx_receipt(receipt, web3)

    def count_queued_sync(
        self,
        chain_id: StarknetChainId,
        web3: Web3,
        block_number: Optional[EthBlockIdentifier] = None,
    ) -> int:
        """

        :param chain_id: A `StarknetChainId` (which contains StarkNet core contract deployed)
        :param web3: Web3 instance from web3.py
        :param block_number: Optional. `EthBlockIdentifier` which is a hex address, integer,
                             or one of "latest", "earliest", "pending"
        :return: an integer (0 or 1, 0 meaning not received or a consumed message,
                 and 1 meaning a queued message waiting for consumer)
        """
        return int_from_bytes(
            StarknetL1Contract(chain_id, web3).l1_to_l2_messages(
                self.hash, block_number
            )
        )



from .address import (
    AddressRepresentation,
    Address,
    parse_address,
    compute_address,
    BlockIdentifier,
)
from .transaction import (
    Transaction,
    InvokeFunction,
    Deploy,
    compute_deploy_hash,
    compute_invoke_hash,
    TransactionType,
)
from .chains import chain_from_network, StarknetChainId



from __future__ import annotations

import dataclasses
import sys
import warnings
from dataclasses import dataclass
from typing import (
    List,
    Optional,
    TypeVar,
    Union,
    Dict,
    Collection,
    NamedTuple,
)

from starkware.cairo.lang.compiler.identifier_manager import IdentifierManager
from starkware.starknet.core.os.class_hash import compute_class_hash
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.public.abi_structs import identifier_manager_from_abi
from starkware.starknet.services.api.contract_class import ContractClass
from starkware.starknet.services.api.feeder_gateway.feeder_gateway_client import (
    CastableToHash,
)

from starknet_py.common import create_compiled_contract
from starknet_py.net import AccountClient

from starknet_py.proxy_check import ProxyCheck, ArgentProxyCheck, OpenZeppelinProxyCheck
from starknet_py.net.models import (
    InvokeFunction,
    AddressRepresentation,
    parse_address,
    compute_address,
    compute_invoke_hash,
)
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.transactions.deploy import make_deploy_tx
from starknet_py.utils.crypto.facade import pedersen_hash, Call
from starknet_py.utils.data_transformer import FunctionCallSerializer
from starknet_py.utils.sync import add_sync_methods

from starknet_py.net.client import Client

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


ABI = list
ABIEntry = dict
TSentTransaction = TypeVar("TSentTransaction", bound="SentTransaction")


@dataclass(frozen=True)
class ContractData:
    address: int
    abi: ABI
    identifier_manager: IdentifierManager

    @staticmethod
    def from_abi(address: int, abi: ABI) -> "ContractData":
        return ContractData(
            address=address,
            abi=abi,
            identifier_manager=identifier_manager_from_abi(abi),
        )


@add_sync_methods
@dataclass(frozen=True)
class SentTransaction:
    """
    Dataclass exposing the interface of transaction related to a performed action
    """

    hash: CastableToHash
    _client: Client
    status: Optional[str] = None
    block_number: Optional[int] = None

    async def wait_for_acceptance(
        self: TSentTransaction,
        wait_for_accept: Optional[bool] = False,
        check_interval=5,
    ) -> TSentTransaction:
        """
        Waits for transaction to be accepted on chain. By default, returns when status is ``PENDING`` -
        use ``wait_for_accept`` to wait till ``ACCEPTED`` status.
        Returns a new SentTransaction instance, **does not mutate original instance**.
        """
        block_number, status = await self._client.wait_for_tx(
            self.hash,
            wait_for_accept=wait_for_accept,
            check_interval=check_interval,
        )
        return dataclasses.replace(
            self,
            status=status,
            block_number=block_number,
        )


@add_sync_methods
@dataclass(frozen=True)
class InvokeResult(SentTransaction):
    contract: ContractData = None
    invoke_transaction: InvokeFunction = None

    def __post_init__(self):
        assert self.contract is not None
        assert self.invoke_transaction is not None


InvocationResult = InvokeResult


@add_sync_methods
@dataclass(frozen=True)
class DeployResult(SentTransaction):
    deployed_contract: "Contract" = None

    def __post_init__(self):
        assert self.deployed_contract is not None


# pylint: disable=too-many-instance-attributes
@add_sync_methods
class PreparedFunctionCall(Call):
    def __init__(
        self,
        calldata: List[int],
        arguments: Dict[str, List[int]],
        selector: int,
        client: Client,
        payload_transformer: FunctionCallSerializer,
        contract_data: ContractData,
        max_fee: int,
        version: int,
    ):
        # pylint: disable=too-many-arguments
        super().__init__(
            to_addr=contract_data.address, selector=selector, calldata=calldata
        )
        self.arguments = arguments
        self._client = client
        self._payload_transformer = payload_transformer
        self._contract_data = contract_data
        self.max_fee = max_fee
        self.version = version

    @property
    def hash(self) -> int:
        warnings.warn("Hash is deprecated and will be deleted in next releases")

        return compute_invoke_hash(
            contract_address=self._contract_data.address,
            entry_point_selector=self.selector,
            calldata=self.calldata,
            chain_id=self._client.chain,
            max_fee=self.max_fee if self.max_fee is not None else 0,
            version=self.version,
        )

    async def call_raw(
        self,
        signature: Optional[Collection[int]] = None,
        block_hash: Optional[str] = None,
    ) -> List[int]:
        """
        Calls a method without translating the result into python values.

        :param signature: Signature to send
        :param block_hash: Optional block hash
        :return: list of ints
        """
        tx = self._make_invoke_function(signature)
        return await self._client.call_contract(invoke_tx=tx, block_hash=block_hash)

    async def call(
        self,
        signature: Optional[Collection[int]] = None,
        block_hash: Optional[str] = None,
    ) -> NamedTuple:
        """
        Calls a method.

        :param signature: Signature to send
        :param block_hash: Optional block hash
        :return: CallResult or List[int] if return_raw is used
        """
        result = await self.call_raw(signature=signature, block_hash=block_hash)
        return self._payload_transformer.to_python(result)

    async def invoke(
        self,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> InvokeResult:
        """
        Invokes a method.

        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :return: InvokeResult
        """
        self._assert_can_invoke()

        if max_fee is not None:
            self.max_fee = max_fee

        transaction = await self._client.sign_transaction(
            self, self.max_fee, auto_estimate, self.version
        )
        response = await self._client.send_transaction(transaction)

        invoke_result = InvokeResult(
            hash=response.transaction_hash,  # noinspection PyTypeChecker
            _client=self._client,
            contract=self._contract_data,
            invoke_transaction=transaction,
        )

        return invoke_result

    async def estimate_fee(self):
        """
        Estimate fee for prepared function call

        :return: Estimated amount of Wei executing specified transaction will cost
        :raises ValueError: when max_fee of PreparedFunctionCall is not None or 0.
        """
        self._assert_can_invoke()

        if self.max_fee is not None and self.max_fee != 0:
            raise ValueError(
                "Cannot estimate fee of PreparedFunctionCall with max_fee not None or 0."
            )

        tx = await self._client.sign_transaction(self, max_fee=0, version=0)

        return await self._client.estimate_fee(tx=tx)

    def _make_invoke_function(self, signature) -> InvokeFunction:
        return InvokeFunction(
            contract_address=self._contract_data.address,
            entry_point_selector=self.selector,
            calldata=self.calldata,
            # List is required here
            signature=[*signature] if signature else [],
            max_fee=self.max_fee if self.max_fee is not None else 0,
            version=self.version,
        )

    def _assert_can_invoke(self):
        if not isinstance(self._client, AccountClient):
            raise ValueError(
                "Contract uses an account that can't invoke transactions. You need to use AccountClient for that."
            )


@add_sync_methods
class ContractFunction:
    def __init__(
        self, name: str, abi: ABIEntry, contract_data: ContractData, client: Client
    ):
        self.name = name
        self.abi = abi
        self.inputs = abi["inputs"]
        self.contract_data = contract_data
        self._client = client
        self._payload_transformer = FunctionCallSerializer(
            abi=self.abi, identifier_manager=self.contract_data.identifier_manager
        )

    def prepare(
        self,
        *args,
        version: int = 0,
        max_fee: int = None,
        **kwargs,
    ) -> PreparedFunctionCall:
        """
        ``*args`` and ``**kwargs`` are translated into Cairo calldata.
         Creates a ``PreparedFunctionCall`` instance
         which exposes calldata for every argument and adds more arguments when calling methods.

        :param version: PreparedFunctionCall version
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :return: PreparedFunctionCall
        """
        calldata, arguments = self._payload_transformer.from_python(*args, **kwargs)
        return PreparedFunctionCall(
            calldata=calldata,
            arguments=arguments,
            contract_data=self.contract_data,
            client=self._client,
            payload_transformer=self._payload_transformer,
            selector=self.get_selector(self.name),
            max_fee=max_fee,
            version=version,
        )

    async def call(
        self,
        *args,
        block_hash: Optional[str] = None,
        **kwargs,
    ) -> NamedTuple:
        """
        :param block_hash: Block hash to execute the contract at specific point of time

        Call contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        The result is translated from Cairo data to python values.
        Equivalent of ``.prepare(*args, **kwargs).call()``.
        """
        return await self.prepare(max_fee=0, version=0, *args, **kwargs).call(
            block_hash=block_hash
        )

    async def invoke(
        self, *args, max_fee: int = None, auto_estimate: bool = False, **kwargs
    ) -> InvokeResult:
        """
        Invoke contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        Equivalent of ``.prepare(*args, **kwargs).invoke()``.

        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        """
        prepared_call = self.prepare(*args, **kwargs)
        return await prepared_call.invoke(max_fee=max_fee, auto_estimate=auto_estimate)

    @staticmethod
    def get_selector(function_name: str):
        """
        :param function_name: Contract function's name
        :return: A StarkNet integer selector for this function inside the contract
        """
        return get_selector_from_name(function_name)


FunctionsRepository = Dict[str, ContractFunction]


@add_sync_methods
class Contract:
    """
    Cairo contract's model.
    """

    def __init__(self, address: AddressRepresentation, abi: list, client: Client):
        """
        Should be used instead of ``from_address`` when ABI is known statically.

        :param address: contract's address
        :param abi: contract's abi
        :param client: client used for API calls
        """
        self.data = ContractData.from_abi(parse_address(address), abi)
        self._functions = self._make_functions(self.data, client)
        self.client = client

    @property
    def functions(self) -> FunctionsRepository:
        """
        :return: All functions exposed from a contract.
        """
        return self._functions

    @property
    def address(self) -> int:
        return self.data.address

    class ProxyConfig(TypedDict, total=False):
        """
        Proxy resolving configuration
        """

        max_steps: int
        """
        Max number of contracts that `Contract.from_address` will process before raising `RecursionError`.
        """
        proxy_checks: List[ProxyCheck]
        """
        List of classes implementing :class:`starknet_py.proxy_check.ProxyCheck` ABC,
        that will be used for checking if contract at the address is a proxy contract.
        """

    @staticmethod
    async def from_address(
        address: AddressRepresentation,
        client: Client,
        proxy_config: Union[bool, ProxyConfig] = False,
    ) -> "Contract":
        """
        Fetches ABI for given contract and creates a new Contract instance with it. If you know ABI statically you
        should create Contract's instances directly instead of using this function to avoid unnecessary API calls.

        :raises ContractNotFoundError: when contract is not found
        :param address: Contract's address
        :param client: Client
        :param proxy_config: Proxy resolving config
            If set to ``True``, will use default proxy checks and :class:
            `starknet_py.proxy_check.OpenZeppelinProxyCheck`
            and :class:`starknet_py.proxy_check.ArgentProxyCheck` and default max_steps = 5.

            If set to ``False``, :meth:`Contract.from_address` will not resolve proxies.

            If a valid `ProxyConfig` is provided, will use values from that instead supplementing
            with defaults when needed.

        :return: an initialized Contract instance
        """
        default_config: Contract.ProxyConfig = {
            "max_steps": 5,
            "proxy_checks": [ArgentProxyCheck(), OpenZeppelinProxyCheck()],
        }
        if isinstance(proxy_config, bool):
            proxy_config = default_config if proxy_config else {}
        else:
            proxy_config = {**default_config, **proxy_config}

        contract = await ContractFromAddressFactory(
            address=address,
            client=client,
            max_steps=proxy_config.get("max_steps", 1),
            proxy_checks=proxy_config.get("proxy_checks", []),
        ).make_contract()

        return Contract(address=address, abi=contract.data.abi, client=client)

    @staticmethod
    async def deploy(
        client: Client,
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        constructor_args: Optional[Union[List[any], dict]] = None,
        salt: Optional[int] = None,
        search_paths: Optional[List[str]] = None,
    ) -> "DeployResult":
        # pylint: disable=too-many-arguments
        """
        Deploys a contract and waits until it has ``PENDING`` status.
        Either `compilation_source` or `compiled_contract` is required.

        :param client: Client
        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param salt: Optional salt. Random value is selected if it is not provided.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.
        :return: DeployResult instance
        """
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        translated_args = Contract._translate_constructor_args(
            compiled_contract, constructor_args
        )
        deploy_tx = make_deploy_tx(
            compiled_contract=compiled_contract,
            constructor_calldata=translated_args,
            salt=salt,
        )
        res = await client.deploy(deploy_tx)
        contract_address = res.contract_address

        deployed_contract = Contract(
            client=client,
            address=contract_address,
            abi=compiled_contract.abi,
        )
        deploy_result = DeployResult(
            hash=res.transaction_hash,
            _client=client,
            deployed_contract=deployed_contract,
        )

        return deploy_result

    @staticmethod
    def compute_address(
        salt: int,
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        constructor_args: Optional[Union[List[any], dict]] = None,
        search_paths: Optional[List[str]] = None,
    ) -> int:
        """
        Computes address for given contract.
        Either `compilation_source` or `compiled_contract` is required.

        :param salt: int
        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.
        :return: contract's address
        """
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        translated_args = Contract._translate_constructor_args(
            compiled_contract, constructor_args
        )
        return compute_address(
            salt=salt,
            contract_hash=compute_class_hash(
                compiled_contract, hash_func=pedersen_hash
            ),
            constructor_calldata=translated_args,
        )

    @staticmethod
    def compute_contract_hash(
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        search_paths: Optional[List[str]] = None,
    ) -> int:
        """
        Computes hash for given contract.
        Either `compilation_source` or `compiled_contract` is required.

        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.
        :return:
        """
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        return compute_class_hash(compiled_contract, hash_func=pedersen_hash)

    @staticmethod
    def _translate_constructor_args(
        contract: ContractClass, constructor_args: any
    ) -> List[int]:
        constructor_abi = next(
            (member for member in contract.abi if member["type"] == "constructor"),
            None,
        )

        # Constructor might not accept any arguments
        if not constructor_abi or not constructor_abi["inputs"]:
            return []

        if not constructor_args:
            raise ValueError(
                "Provided contract has a constructor and no args were provided."
            )

        args, kwargs = (
            ([], constructor_args)
            if isinstance(constructor_args, dict)
            else (constructor_args, {})
        )
        calldata, _args = FunctionCallSerializer(
            constructor_abi, identifier_manager_from_abi(contract.abi)
        ).from_python(*args, **kwargs)
        return calldata

    @classmethod
    def _make_functions(
        cls, contract_data: ContractData, client: Client
    ) -> FunctionsRepository:
        repository = {}

        for abi_entry in contract_data.abi:
            if abi_entry["type"] != "function":
                continue

            name = abi_entry["name"]
            repository[name] = ContractFunction(
                name=name,
                abi=abi_entry,
                contract_data=contract_data,
                client=client,
            )

        return repository


class ContractFromAddressFactory:
    def __init__(
        self,
        address: AddressRepresentation,
        client: Client,
        max_steps: int,
        proxy_checks: List[ProxyCheck],
    ):
        self._address = address
        self._client = client
        self._max_steps = max_steps
        self._proxy_checks = proxy_checks
        self._processed_addresses = set()

    async def make_contract(self) -> Contract:
        return await self._make_contract_recursively(step=1, address=self._address)

    async def _make_contract_recursively(
        self, step: int, address: AddressRepresentation
    ) -> Contract:
        if address in self._processed_addresses:
            raise RecursionError("Proxy cycle detected while resolving proxies")

        if step > self._max_steps:
            raise RecursionError("Max number of steps exceeded while resolving proxies")

        code = await self._client.get_code(contract_address=parse_address(address))
        contract = Contract(
            address=parse_address(address), abi=code.abi, client=self._client
        )
        self._processed_addresses.add(address)

        is_proxy = False
        address = 0
        for proxy_check in self._proxy_checks:
            if await proxy_check.is_proxy(contract):
                is_proxy = True
                address = await proxy_check.implementation_address(contract)
                break

        if not is_proxy:
            return contract

        return await self._make_contract_recursively(
            address=address,
            step=step + 1,
        )



from abc import ABC, abstractmethod
from typing import List

from starknet_py.net.models.transaction import Transaction


class BaseSigner(ABC):
    """
    Base class for transaction signer. Implement methods from this ABC to use a custom signer in AccountClient
    """

    @property
    @abstractmethod
    def public_key(self) -> int:
        """
        Public key of the signer

        :return: public key
        """

    @abstractmethod
    def sign_transaction(
        self,
        transaction: Transaction,
    ) -> List[int]:
        """
        Sign execute transaction and return a signature

        :param transaction: Execute transaction to sign
        :return: transaction signature
        """



from typing import Union, List, Optional

from starkware.starknet.definitions.fields import ContractAddressSalt

from starknet_py.common import create_compiled_contract
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.net.client_models import ContractClass
from starknet_py.net.models import Deploy


# pylint: disable=too-many-arguments
def make_deploy_tx(
    compilation_source: Optional[StarknetCompilationSource] = None,
    compiled_contract: Optional[Union[str, ContractClass]] = None,
    constructor_calldata: List[int] = None,
    salt: Optional[int] = None,
    version: int = 0,
    cairo_path: Optional[List[str]] = None,
) -> Deploy:
    """
    Create deploy tx.
    Either `compilation_source` or `compiled_contract` is required.

    :param compilation_source: string containing source code or a list of source files paths
    :param compiled_contract: string containing compiled contract bytecode.
                              Useful for reading compiled contract from a file
    :param constructor_calldata: Calldata to be passed to contract constructor
    :param salt: Salt to be used when generating contract address
    :param version: PreparedFunctionCall version
    :param cairo_path: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts
    :return: A "Deploy" transaction object
    """
    if not constructor_calldata:
        constructor_calldata = []

    if isinstance(compiled_contract, str):
        compiled_contract = ContractClass.loads(compiled_contract)

    if not compiled_contract:
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, cairo_path
        )

    return Deploy(
        contract_address_salt=ContractAddressSalt.get_random_value()
        if salt is None
        else salt,
        contract_definition=compiled_contract,
        constructor_calldata=constructor_calldata,
        version=version,
    )



import json
from typing import Tuple, List

from marshmallow import Schema, fields, post_load, pre_load, EXCLUDE
from marshmallow_oneofschema import OneOfSchema

from starknet_py.net.client_models import (
    ContractCode,
    StarknetBlock,
    L1toL2Message,
    L2toL1Message,
    BlockStateUpdate,
    StorageDiff,
    ContractDiff,
    Event,
    EntryPoint,
    EntryPointsByType,
    DeclaredContract,
    InvokeTransaction,
    DeclareTransaction,
    DeployTransaction,
    TransactionReceipt,
    SentTransactionResponse,
    DeclareTransactionResponse,
    DeployTransactionResponse,
)
from starknet_py.net.common_schemas.common_schemas import (
    Felt,
    BlockStatusField,
    StatusField,
    NonPrefixedHex,
)

# pylint: disable=unused-argument
# pylint: disable=no-self-use


class FunctionCallSchema(Schema):
    contract_address = fields.Integer(data_key="contract_address", required=True)
    entry_point_selector = fields.Integer(
        data_key="entry_point_selector", required=True
    )
    calldata = fields.List(fields.Integer(), data_key="calldata", required=True)


class EventSchema(Schema):
    from_address = Felt(data_key="from_address", required=True)
    keys = fields.List(Felt(), data_key="keys", required=True)
    data = fields.List(Felt(), data_key="data", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Event:
        return Event(**data)


class L1toL2MessageSchema(Schema):
    l1_address = Felt(data_key="from_address", required=True)
    l2_address = Felt(load_default=None)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L1toL2Message:
        return L1toL2Message(**data)


class L2toL1MessageSchema(Schema):
    l2_address = Felt(load_default=None)
    l1_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L2toL1Message:
        return L2toL1Message(**data)


class TransactionReceiptSchema(Schema):
    hash = Felt(data_key="txn_hash", required=True)
    status = StatusField(data_key="status", required=True)
    actual_fee = Felt(data_key="actual_fee", required=True)
    rejection_reason = fields.String(data_key="statusData", load_default=None)
    events = fields.List(
        fields.Nested(EventSchema()), data_key="events", load_default=[]
    )
    l1_to_l2_consumed_message = fields.Nested(
        L1toL2MessageSchema(), data_key="l1_origin_message", load_default=None
    )
    l2_to_l1_messages = fields.List(
        fields.Nested(L2toL1MessageSchema()), data_key="messages_sent", load_default=[]
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionReceipt:
        return TransactionReceipt(**data)


class ContractCodeSchema(Schema):
    bytecode = fields.List(Felt(), data_key="bytecode", required=True)
    abi = fields.String(data_key="abi", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> ContractCode:
        data["abi"] = ContractCodeSchema._abi_to_dict(data["abi"])
        return ContractCode(**data)

    @staticmethod
    def _abi_to_dict(abi: str) -> dict:
        return json.loads(abi)


class TransactionSchema(Schema):
    hash = Felt(data_key="txn_hash", required=True)
    signature = fields.List(Felt(), data_key="signature", load_default=[])
    max_fee = Felt(data_key="max_fee", load_default=0)


class InvokeTransactionSchema(TransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)
    calldata = fields.List(Felt(), data_key="calldata", required=True)

    @pre_load
    def preprocess(self, data, **kwargs):
        if data["calldata"] is None:
            data["calldata"] = []
        return data

    @post_load
    def make_transaction(self, data, **kwargs) -> InvokeTransaction:
        return InvokeTransaction(**data)


class DeclareTransactionSchema(TransactionSchema):
    class_hash = Felt(data_key="class_hash", required=True)
    sender_address = Felt(data_key="sender_address", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransaction:
        return DeclareTransaction(**data)


class DeployTransactionSchema(TransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)
    constructor_calldata = fields.List(Felt(), data_key="calldata", required=True)
    # class_hash = Felt(data_key="class_hash", load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployTransaction:
        return DeployTransaction(**data)


class TypesOfTransactionsSchema(OneOfSchema):
    type_field = "type"
    type_schemas = {
        "INVOKE_FUNCTION": InvokeTransactionSchema,
        "DECLARE": DeclareTransactionSchema,
        "DEPLOY": DeployTransactionSchema,
    }

    def get_data_type(self, data):
        if "contract_class" in data:
            return "DECLARE"
        if "entry_point_selector" in data and data["entry_point_selector"] is not None:
            return "INVOKE_FUNCTION"
        return "DEPLOY"


class StarknetBlockSchema(Schema):
    block_hash = Felt(data_key="block_hash", required=True)
    parent_block_hash = Felt(data_key="parent_hash", required=True)
    block_number = fields.Integer(data_key="block_number", required=True)
    status = BlockStatusField(data_key="status", required=True)
    root = NonPrefixedHex(data_key="new_root", required=True)
    transactions = fields.List(
        fields.Nested(TypesOfTransactionsSchema(unknown=EXCLUDE)),
        data_key="transactions",
        required=True,
    )
    timestamp = fields.Integer(data_key="timestamp", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> StarknetBlock:
        return StarknetBlock(**data)


class StorageDiffSchema(Schema):
    address = Felt(data_key="address", required=True)
    key = Felt(data_key="key", required=True)
    value = Felt(data_key="value", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> StorageDiff:
        return StorageDiff(**data)


class ContractDiffSchema(Schema):
    address = Felt(data_key="address", required=True)
    contract_hash = Felt(data_key="contract_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> ContractDiff:
        return ContractDiff(**data)


class StateDiffSchema(Schema):
    storage_diffs = fields.List(
        fields.Nested(StorageDiffSchema()), data_key="storage_diffs", required=True
    )
    contract_diffs = fields.List(
        fields.Nested(ContractDiffSchema()), data_key="contracts", required=True
    )


class BlockStateUpdateSchema(Schema):
    block_hash = Felt(data_key="block_hash", required=True)
    new_root = Felt(data_key="new_root", required=True)
    old_root = Felt(data_key="old_root", required=True)
    state_diff = fields.Nested(StateDiffSchema(), data_key="state_diff", required=True)

    @pre_load
    def preprocess(self, data, **kwargs):
        # Remove this when support for nonces is added
        del data["state_diff"]["nonces"]

        return data

    @post_load
    def make_dataclass(self, data, **kwargs) -> BlockStateUpdate:
        storage_diffs, contract_diffs = BlockStateUpdateSchema._extract_diffs(
            data["state_diff"]
        )
        del data["state_diff"]

        return BlockStateUpdate(
            **data,
            storage_diffs=storage_diffs,
            contract_diffs=contract_diffs,
            declared_contracts=[],
        )

    @staticmethod
    def _extract_diffs(
        state_diff: dict,
    ) -> Tuple[List[StorageDiff], List[ContractDiff]]:
        storage_diffs = state_diff["storage_diffs"]
        contract_diffs = state_diff["contract_diffs"]
        return storage_diffs, contract_diffs


class EntryPointSchema(Schema):
    offset = Felt(data_key="offset", required=True)
    selector = Felt(data_key="selector", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPoint:
        return EntryPoint(**data)


class EntryPointsByTypeSchema(Schema):
    constructor = fields.List(
        fields.Nested(EntryPointSchema()), data_key="CONSTRUCTOR", required=True
    )
    external = fields.List(
        fields.Nested(EntryPointSchema()), data_key="EXTERNAL", required=True
    )
    l1_handler = fields.List(
        fields.Nested(EntryPointSchema()), data_key="L1_HANDLER", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPointsByType:
        return EntryPointsByType(**data)


class DeclaredContractSchema(Schema):
    program = fields.String(data_key="program", required=True)
    entry_points_by_type = fields.Nested(
        EntryPointsByTypeSchema(), data_key="entry_points_by_type", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclaredContract:
        return DeclaredContract(**data)


class SentTransactionSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return SentTransactionResponse(**data)


class DeclareTransactionResponseSchema(SentTransactionSchema):
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return DeclareTransactionResponse(**data)


class DeployTransactionResponseSchema(SentTransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return DeployTransactionResponse(**data)



from typing import Optional, List

from starkware.starknet.services.api.gateway.transaction import DECLARE_SENDER_ADDRESS

from starknet_py.common import create_compiled_contract
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.net.models.transaction import Declare


def make_declare_tx(
    compilation_source: Optional[StarknetCompilationSource] = None,
    compiled_contract: Optional[str] = None,
    version: int = 0,
    cairo_path: Optional[List[str]] = None,
) -> Declare:
    """
    Create declaration tx.
    Either `compilation_source` or `compiled_contract` is required.

    :param compilation_source: string containing source code or a list of source files paths
    :param compiled_contract: string containing compiled contract bytecode.
                              Useful for reading compiled contract from a file
    :param version: PreparedFunctionCall version
    :param cairo_path: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts
    :return: A "Declare" transaction object
    """
    compiled_contract = create_compiled_contract(
        compilation_source, compiled_contract, cairo_path
    )
    return Declare(
        contract_class=compiled_contract,
        sender_address=DECLARE_SENDER_ADDRESS,
        max_fee=0,
        signature=[],
        nonce=0,
        version=version,
    )



from starkware.starknet.public.abi_structs import identifier_manager_from_abi

from starknet_py.utils.data_transformer import FunctionCallSerializer

abi = [
    {
        "inputs": [
            {"name": "call_array_len", "type": "felt"},
            {"name": "call_array", "type": "CallArray*"},
            {"name": "calldata_len", "type": "felt"},
            {"name": "calldata", "type": "felt*"},
            {"name": "nonce", "type": "felt"},
        ],
        "name": "__execute__",
        "outputs": [
            {"name": "retdata_size", "type": "felt"},
            {"name": "retdata", "type": "felt*"},
        ],
        "type": "function",
    },
    {
        "members": [
            {"name": "to", "offset": 0, "type": "felt"},
            {"name": "selector", "offset": 1, "type": "felt"},
            {"name": "data_offset", "offset": 2, "type": "felt"},
            {"name": "data_len", "offset": 3, "type": "felt"},
        ],
        "name": "CallArray",
        "size": 4,
        "type": "struct",
    },
]

execute_transformer = FunctionCallSerializer(
    abi=abi[0], identifier_manager=identifier_manager_from_abi(abi)
)



import inspect
import os
import sys
from typing import TypeVar

generating_docs = os.path.basename(sys.argv[0]) == "sphinx-build"
T = TypeVar("T")


def as_our_module(cls: T) -> T:
    """
    We create alias for some function from starkware, but sphinx show original module path. This makes docs harder to
    follow.

    :param cls: class
    :return: Same class, just with changed module if docs are being created.
    """
    if not generating_docs:
        return cls

    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    cls.__module__ = mod.__name__
    return cls



from typing import Any, Union, Mapping, Optional

from marshmallow import fields, ValidationError

from starknet_py.net.client_models import (
    TransactionStatus,
    BlockStatus,
    TransactionType,
)

# pylint: disable=unused-argument


class Felt(fields.Field):
    """
    Field that serializes int to felt (hex encoded string)
    """

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return hex(value)

    def _deserialize(
        self,
        value: Any,
        attr: Union[str, None],
        data: Union[Mapping[str, Any], None],
        **kwargs,
    ):
        if not isinstance(value, str) or not value.startswith("0x"):
            raise ValidationError(f"Invalid value provided for felt: {value}")

        try:
            return int(value, 16)
        except ValueError as error:
            raise ValidationError("Invalid felt") from error


class NonPrefixedHex(fields.Field):
    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return hex(value).lstrip("0x")

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        return int(value, 16)


class StatusField(fields.Field):
    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionStatus:
        values = [v.value for v in TransactionStatus]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionStatus: {value}"
            )

        return TransactionStatus(value)


class BlockStatusField(fields.Field):
    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> BlockStatus:
        values = [v.value for v in BlockStatus]

        if value in ("ABORTED", "REVERTED"):
            return BlockStatus.REJECTED

        if value not in values:
            raise ValidationError(f"Invalid value for BlockStatus provided: {value}")

        return BlockStatus(value)


class TransactionTypeField(fields.Field):
    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionType:
        values = [v.value for v in TransactionType]

        if value == "INVOKE_FUNCTION":
            value = "INVOKE"

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionType: {value}"
            )

        return TransactionType(value)



import os
from typing import Callable, Iterable, Optional
from dataclasses import dataclass

from starkware.cairo.common.hash_state import compute_hash_on_elements
from starkware.cairo.lang.vm.crypto import pedersen_hash as default_hash
from starkware.crypto.signature.signature import sign

from crypto_cpp_py.cpp_bindings import (
    cpp_hash,
    get_cpp_lib_file,
    ECSignature,
)
from starknet_py.net.client_models import Call


# PREFIX_TRANSACTION = 'StarkNet Transaction'
PREFIX_TRANSACTION = 476441609247967894954472788179128007176248455022


@dataclass(frozen=True)
class MultiCall:
    account: int
    calls: Iterable[Call]
    nonce: int
    max_fee: int = 0
    version: int = 0


# pylint: disable=too-many-arguments
def hash_multicall_with(
    multi_call: MultiCall,
    hash_fun: Callable[[int, int], int],
) -> int:
    """
    Mimics the behavior of
    https://github.com/argentlabs/cairo-contracts/blob/c2ff198e5de5b19514d99ecff604a7cbf3377d2f/contracts/Account.cairo#L248
    """

    calls_hash = compute_hash_on_elements(
        [hash_call_with(c, hash_fun=hash_fun) for c in multi_call.calls],
        hash_func=hash_fun,
    )

    return compute_hash_on_elements(
        [
            PREFIX_TRANSACTION,
            multi_call.account,
            calls_hash,
            multi_call.nonce,
            multi_call.max_fee,
            multi_call.version,
        ],
        hash_func=hash_fun,
    )


def hash_call_with(call: Call, hash_fun):
    return compute_hash_on_elements(
        [
            call.to_addr,
            call.selector,
            compute_hash_on_elements(
                call.calldata,
                hash_func=hash_fun,
            ),
        ]
    )


# Interface
def use_cpp_variant() -> bool:
    force_disable_ext = (
        os.getenv("DISABLE_CRYPTO_C_EXTENSION", "false").lower() == "true"
    )
    cpp_lib_file = get_cpp_lib_file()
    return not force_disable_ext and bool(cpp_lib_file)


def message_signature(msg_hash, priv_key, seed: Optional[int] = 32) -> ECSignature:
    return sign(msg_hash, priv_key, seed)


def pedersen_hash(left: int, right: int) -> int:
    if use_cpp_variant():
        return cpp_hash(left, right)
    return default_hash(left, right)


def hash_multicall(multi_call: MultiCall) -> int:
    return hash_multicall_with(
        multi_call=multi_call,
        hash_fun=pedersen_hash,
    )
